#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s module
"""
from legion.k8s.enclave import Enclave, find_enclaves
from legion.k8s.services import Service, ModelService
from legion.k8s.definitions import ModelIdVersion, ModelDeploymentDescription
from legion.k8s.definitions import STATUS_OK, STATUS_FAIL, STATUS_WARN
from legion.k8s.definitions import EVENT_ADDED, EVENT_MODIFIED, EVENT_DELETED
from legion.k8s.utils import build_client, get_current_namespace, load_config, load_secrets
from legion.k8s.utils import CONNECTION_CONTEXT
from legion.k8s.watch import ResourceWatch
from legion.k8s.properties import K8SPropertyStorage, K8SConfigMapStorage, K8SSecretStorage



#!/usr/bin/env python
#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Logging controller
"""

import logging
import os
import sys

import legion.config
import legion.utils

ROOT_LOGGER = logging.getLogger()


def redirect_to_stdout():
    """
    Redirect ROOT log output to stdout
    :return: None
    """
    channel = logging.StreamHandler(sys.stderr)
    channel.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    channel.setFormatter(formatter)
    ROOT_LOGGER.addHandler(channel)


def set_log_level(target=None):
    """
    Update log level to target or to value from ENV
    :param target: target log level (level from logging)
    :type target: str
    :return: None
    """
    if target:
        log_level = target
    else:
        if legion.utils.string_to_bool(os.getenv('VERBOSE', '')):
            log_level = logging.DEBUG
        else:
            log_level = logging.ERROR

    ROOT_LOGGER.setLevel(log_level)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s definitions functions
"""
import typing

ModelDeploymentDescription = typing.NamedTuple('ModelDeploymentDescription', [
    ('status', str),
    ('model', str),
    ('version', str),
    ('image', str),
    ('scale', int),
    ('ready_replicas', int),
    ('namespace', str),
    ('model_api_ok', bool),
    ('model_api_info', dict),
])
ModelIdVersion = typing.NamedTuple('ModelDeploymentDescription', [
    ('id', str),
    ('version', str),
])


LEGION_SYSTEM_LABEL = 'legion.system'
LEGION_SYSTEM_VALUE = 'yes'
LEGION_COMPONENT_LABEL = 'legion.component'
LEGION_COMPONENT_NAME_MODEL = 'model'
LEGION_COMPONENT_NAME_EDI = 'edi'
LEGION_COMPONENT_NAME_API = 'edge'
LEGION_COMPONENT_NAME_GRAFANA = 'grafana'
LEGION_COMPONENT_NAME_GRAPHITE = 'graphite'

LEGION_API_SERVICE_PORT = 'api'

ENCLAVE_NAMESPACE_LABEL = 'enclave'

INGRESS_DASHBOARD_NAME = 'kubernetes-dashboard'
INGRESS_GRAFANA_NAME = 'legion-core-grafana'

STATUS_OK = 'ok'
STATUS_FAIL = 'fail'
STATUS_WARN = 'warning'

EVENT_ADDED = 'ADDED'
EVENT_MODIFIED = 'MODIFIED'
EVENT_DELETED = 'DELETED'



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion headers (for meta-information)
"""

STDERR_PREFIX = 'X-Legion-'

MODEL_ID = 'Model-Id'
MODEL_PATH = 'Model-Path'
MODEL_VERSION = 'Model-Version'
IMAGE_TAG_EXTERNAL = 'Model-Image-Tag-External'
IMAGE_TAG_LOCAL = 'Model-Image-Tag-Local'
SAVE_STATUS = 'Save-Status'

DOMAIN_PREFIX = 'com.epam.'
DOMAIN_MODEL_ID = DOMAIN_PREFIX + 'legion.model.id'
DOMAIN_MODEL_VERSION = DOMAIN_PREFIX + 'legion.model.version'
DOMAIN_CLASS = DOMAIN_PREFIX + 'legion.class'
DOMAIN_CONTAINER_TYPE = DOMAIN_PREFIX + 'legion.container_type'
DOMAIN_CONTAINER_DESCRIPTION = DOMAIN_PREFIX + 'legion.container_description'
DOMAIN_MODEL_PROPERTIES = DOMAIN_PREFIX + 'legion.model.properties'
DOMAIN_MODEL_PROPERTY_VALUES = DOMAIN_PREFIX + 'legion.model.property_values'
DOMAIN_MODEL_PROPERTY_TYPE = DOMAIN_PREFIX + 'legion.property'



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Flask app
"""

import logging

import legion.config
import legion.k8s
import legion.external.grafana
import legion.http
import legion.model
from flask import Flask, Blueprint
from flask import current_app as app

LOGGER = logging.getLogger(__name__)
blueprint = Blueprint('apiserver', __name__)

EDI_VERSION = '1.0'
EDI_ROOT = '/'
EDI_DEPLOY = '/api/{version}/deploy'
EDI_UNDEPLOY = '/api/{version}/undeploy'
EDI_SCALE = '/api/{version}/scale'
EDI_INSPECT = '/api/{version}/inspect'
EDI_INFO = '/api/{version}/info'

ALL_EDI_API_ENDPOINTS = EDI_DEPLOY, EDI_UNDEPLOY, EDI_SCALE, EDI_INSPECT, EDI_INFO


def build_blueprint_url(endpoint_url_template):
    """
    Build endpoint url from template

    :param endpoint_url_template: endpoint url template
    :type endpoint_url_template: str
    :return: str -- builded url
    """
    return endpoint_url_template.format(version=EDI_VERSION)


def authenticate(user, password):
    """
    Authenticate validator. For token auth use 'token' as user and token value as password

    :param user: user or 'token' (for token auth)
    :type user: str or None
    :param password: password or token value
    :type password: str or None
    :raises Exception: if something wrong with authentication process
    :return: bool -- is user authenticated for performing any operations
    """
    if not app.config['AUTH_ENABLED']:
        return True

    if user == 'token':
        if app.config['AUTH_TOKEN_ENABLED']:
            return password == app.config['AUTH_TOKEN']

        return False

    # TODO: Add LDAP authorisation
    return False


@blueprint.route(build_blueprint_url(EDI_ROOT), methods=['GET'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
def root():
    """
    Root API endpoint

    :return: dict -- root information
    """
    return {
        'component': 'EDI server',
        'version': EDI_VERSION,
        'endpoints': [build_blueprint_url(url) for url in ALL_EDI_API_ENDPOINTS]
    }


@blueprint.route(build_blueprint_url(EDI_DEPLOY), methods=['POST'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
@legion.http.populate_fields(image=str, count=int)
@legion.http.requested_fields('image')
def deploy(image, count=1):
    """
    Deploy API endpoint

    :param image: Docker image for deploy (for jybernetes deployment and local pull)
    :type image: str
    :param count: count of pods to create
    :type count: int
    :return: bool -- True
    """
    LOGGER.info('Command: deploy image {} with {} replicas'.format(image, count))
    model_service = app.config['ENCLAVE'].deploy_model(image, count)

    if app.config['REGISTER_ON_GRAFANA']:
        if not app.config['GRAFANA_CLIENT'].is_dashboard_exists(model_service.id, model_service.version):
            app.config['GRAFANA_CLIENT'].create_dashboard_for_model(model_service.id, model_service.version)
    return True


@blueprint.route(build_blueprint_url(EDI_UNDEPLOY), methods=['POST'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
@legion.http.populate_fields(model=str, version=str, grace_period=int)
@legion.http.requested_fields('model')
def undeploy(model, version=None, grace_period=0):
    """
    Undeploy API endpoint

    :param model: model id
    :type model: str
    :param version: (Optional) specific model version
    :type version: str or None
    :param grace_period: grace period for removing
    :type grace_period: int
    :return: bool -- True
    """
    # TODO: Add tests for multiple versions (parallel models)
    LOGGER.info('Command: undeploy model with id={}, version={} with grace period {}s'
                .format(model, version, grace_period))
    model_services = app.config['ENCLAVE'].get_models_strict(model, version)

    for model_service in model_services:
        if app.config['REGISTER_ON_GRAFANA']:
            if app.config['GRAFANA_CLIENT'].is_dashboard_exists(model, model_service.version):
                app.config['GRAFANA_CLIENT'].remove_dashboard_for_model(model, model_service.version)

        model_service.delete(grace_period)

    return True


@blueprint.route(build_blueprint_url(EDI_SCALE), methods=['POST'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
@legion.http.populate_fields(model=str, count=int, version=str)
@legion.http.requested_fields('model', 'count')
def scale(model, count, version=None):
    """
    Scale API endpoint

    :param model: model id
    :type model: str
    :param count: count of pods to create
    :type count: int
    :param version: (Optional) specific model version
    :type version: str or None
    :return: bool -- True
    """
    LOGGER.info('Command: scale model with id={}, version={} to {} replicas'.format(model, version, count))
    model_services = app.config['ENCLAVE'].get_models_strict(model, version)

    for model_service in model_services:
        model_service.scale = count

    return True


@blueprint.route(build_blueprint_url(EDI_INSPECT), methods=['GET'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
@legion.http.populate_fields(model=str, version=str)
def inspect(model=None, version=None):
    """
    Inspect API endpoint

    :return: dict -- state of cluster models
    """
    LOGGER.info('Command: inspect with model={}, version={}'.format(model, version))
    model_deployments = []
    model_services = app.config['ENCLAVE'].get_models(model, version)

    for model_service in model_services:
        model_api_info = {}

        model_client = legion.model.ModelClient(
            model_id=model_service.id,
            model_version=model_service.version,
            host=model_service.url_with_ip,
            timeout=3
        )
        LOGGER.info('Building model client: {!r}'.format(model_client))

        try:
            model_api_info['result'] = model_client.info()
            model_api_ok = True
        except Exception as model_api_exception:
            LOGGER.error('Cannot connect to model <{}> endpoint to get info: {}'.format(model_service,
                                                                                        model_api_exception))
            model_api_info['exception'] = str(model_api_exception)
            model_api_ok = False

        model_deployments.append(
            legion.k8s.ModelDeploymentDescription(
                status=model_service.status,
                model=model_service.id,
                version=model_service.version,
                image=model_service.image,
                scale=model_service.desired_scale,
                ready_replicas=model_service.scale,
                namespace=model_service.namespace,
                model_api_ok=model_api_ok,
                model_api_info=model_api_info,
            )
        )

    return [x._asdict() for x in model_deployments]


@blueprint.route(build_blueprint_url(EDI_INFO), methods=['GET'])
@legion.http.provide_json_response
@legion.http.authenticate(authenticate)
def info():
    """
    Info API endpoint

    :return: dict -- state of cluster
    """
    return app.config['CLUSTER_STATE']


def create_application():
    """
    Create Flask application and register blueprints

    :return: :py:class:`Flask.app` -- Flask application instance
    """
    application = Flask(__name__, static_url_path='')

    application.register_blueprint(blueprint)

    return application


def load_cluster_config(application):
    """
    Load cluster configuration into Flask config

    :param application: Flask app instance
    :type application: :py:class:`Flask.app`
    :return: None
    """
    if not application.config['NAMESPACE']:
        application.config['NAMESPACE'] = legion.k8s.get_current_namespace()

    application.config['ENCLAVE'] = legion.k8s.Enclave(application.config['NAMESPACE'])

    application.config['CLUSTER_SECRETS'] = legion.k8s.load_secrets(
        application.config['CLUSTER_SECRETS_PATH'])
    application.config['CLUSTER_STATE'] = legion.k8s.load_config(application.config['CLUSTER_CONFIG_PATH'])

    grafana_client = legion.external.grafana.GrafanaClient(application.config['ENCLAVE'].grafana_service.url,
                                                           application.config['CLUSTER_SECRETS']['grafana.user'],
                                                           application.config['CLUSTER_SECRETS']['grafana.password'])
    application.config['GRAFANA_CLIENT'] = grafana_client


def init_application(args=None):
    """
    Initialize configured Flask application instance
    Overall configuration priority: config_default.py, env::FLASK_APP_SETTINGS_FILES file,
    ENV parameters, CLI parameters

    :param args: arguments if provided
    :type args: :py:class:`argparse.Namespace` or None
    :return: :py:class:`Flask.app` -- application instance
    """
    application = create_application()
    legion.http.configure_application(application, args)
    load_cluster_config(application)

    return application


def serve(args):
    """
    Serve apiserver

    :param args: arguments
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    logging.info('Legion api server initializing')
    application = init_application(args)

    try:
        application.run(host=application.config['LEGION_API_ADDR'],
                        port=application.config['LEGION_API_PORT'],
                        debug=application.config['DEBUG'],
                        use_reloader=False)

        return application
    except Exception as run_exception:
        LOGGER.exception('EDI server exited with exception', exc_info=run_exception)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Model HTTP API client and utils
"""

import os
import json
import requests

import legion.config
import legion.http
from legion.utils import normalize_name

from PIL import Image as PYTHON_Image


def load_image(path):
    """
    Load image for model

    :param path: path to local image
    :type path: str
    :return: bytes -- image content
    """
    with PYTHON_Image.open(path) as image:
        if not isinstance(image, PYTHON_Image.Image):
            raise Exception('Invalid image type')

        with open(path, 'rb') as stream:
            return stream.read()


class ModelClient:
    """
    Model HTTP client
    """

    def __init__(self, model_id, model_version, host=None, http_client=None, use_relative_url=False, timeout=None):
        """
        Build client

        :param model_id: model id
        :type model_id: str
        :param model_version: model version
        :type model_version: str
        :param host: host that server model HTTP requests (default: from ENV)
        :type host: str or None
        :param http_client: HTTP client (default: requests)
        :type http_client: python class that implements requests-like post & get methods
        :param use_relative_url: use non-full get/post requests (useful for locust)
        :type use_relative_url: bool
        :param timeout: timeout for connections
        :type timeout: int
        """
        self._model_id = normalize_name(model_id)
        self._model_version = model_version

        if host:
            self._host = host
        else:
            self._host = os.environ.get(*legion.config.MODEL_SERVER_URL)

        if http_client:
            self._http_client = http_client
        else:
            self._http_client = requests

        self._use_relative_url = use_relative_url
        if self._use_relative_url:
            self._host = ''
        else:
            self._host = self._host.rstrip('/')

        self._timeout = timeout

    def __repr__(self):
        """
        Get string representation of model client

        :return: str -- model client information
        """
        return '{}({!r}, {!r}, {!r}, {!r}, {!r})'.format(self.__class__.__name__,
                                                         self._model_id, self._host,
                                                         self._http_client, self._use_relative_url, self._timeout)

    __str__ = __repr__

    @property
    def api_url(self):
        """
        Build API root URL

        :return: str -- api root url
        """
        return '{host}/api/model/{model_id}/{model_version}'.format(host=self._host,
                                                                    model_id=self._model_id,
                                                                    model_version=self._model_version)

    def build_invoke_url(self, endpoint=None):
        """
        Build API invoke URL

        :param endpoint: (Optional) target endpoint
        :type endpoint: str
        :return: str -- invoke url
        """
        if endpoint:
            return '{}/invoke/{}'.format(self.api_url, endpoint)
        return '{}/invoke'.format(self.api_url)

    def build_batch_url(self, endpoint=None):
        """
        Build API batch invoke URL

        :param endpoint: (Optional) target endpoint
        :type endpoint: str
        :return: str -- batch invoke url
        """
        if endpoint:
            return '{}/batch/{}'.format(self.api_url, endpoint)
        return '{}/batch'.format(self.api_url)

    @property
    def info_url(self):
        """
        Build API info URL

        :return: str -- info url
        """
        return self.api_url + '/info'

    @staticmethod
    def _parse_response(response):
        """
        Parse model response (requests or FlaskClient)

        :param response: model HTTP response
        :type response: object with .text or .data and .status_code attributes
        :return: dict -- parsed response
        """
        data = response.text if hasattr(response, 'text') else response.data

        if isinstance(data, bytes):
            data = data.decode('utf-8')

        try:
            data = json.loads(data)
        except ValueError:
            pass

        if not 200 <= response.status_code < 400:
            url = response.url if hasattr(response, 'url') else None
            raise Exception('Wrong status code returned: {}. Data: {}. URL: {}'
                            .format(response.status_code, data, url))

        return data

    @staticmethod
    def _prepare_invoke_request(**parameters):
        """
        Build POST and FILE fields for request due their type

        :param parameters: dict -- invoke parameters
        :return: tuple[dict, dict] -- POST and FILE dictionaries in tuple
        """
        post_fields_dict = {k: v for (k, v) in parameters.items() if not isinstance(v, bytes)}
        post_files = {k: v for (k, v) in parameters.items() if isinstance(v, bytes)}

        post_fields_list = []
        for (k, v) in post_fields_dict.items():
            if isinstance(v, tuple) or isinstance(v, list):
                for item in v:
                    post_fields_list.append((k + '[]', str(item)))
            else:
                post_fields_list.append((k, str(v)))

        return post_fields_list, post_files

    @property
    def _additional_kwargs(self):
        """
        Get additional HTTP client key-value arguments like timestamp

        :return: dict -- additional kwargs
        """
        kwargs = {}
        if self._timeout is not None:
            kwargs['timeout'] = self._timeout
        return kwargs

    def batch(self, invoke_parameters, endpoint=None):
        """
        Send batch invoke request

        :param invoke_parameters: list of dictionaries
        :type invoke_parameters: list[dict]
        :param endpoint: name of endpoint
        :type endpoint: str
        :return: list -- parsed model response
        """
        request_lines = []
        for parameters in invoke_parameters:
            data, files = self._prepare_invoke_request(**parameters)
            if files:
                raise Exception('Files object not allowed for batch invocation')

            request_lines.append(legion.http.encode_http_params(data))

        content = '\n'.join(request_lines)
        url = self.build_batch_url(endpoint)
        response = self._http_client.post(url,
                                          data=content,
                                          **self._additional_kwargs)
        return self._parse_response(response)

    def invoke(self, endpoint=None, **parameters):
        """
        Invoke model with parameters

        :param parameters: parameters for model
        :type parameters: dict[str, object] -- dictionary with parameters
        :param endpoint: name of endpoint
        :type endpoint: str
        :return: dict -- parsed model response
        """
        data, files = self._prepare_invoke_request(**parameters)
        url = self.build_invoke_url(endpoint)
        response = self._http_client.post(url,
                                          data=data, files=files,
                                          **self._additional_kwargs)

        return self._parse_response(response)

    def info(self):
        """
        Get model info

        :return: dict -- parsed model info
        """
        response = self._http_client.get(self.info_url,
                                         **self._additional_kwargs)

        return self._parse_response(response)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s services classes
"""
import logging

import kubernetes
import kubernetes.client
import kubernetes.config
import kubernetes.config.config_exception

import legion.model
from legion.containers.headers import DOMAIN_MODEL_ID, DOMAIN_MODEL_VERSION
import legion.k8s.enclave
import legion.containers.headers
from legion.k8s.definitions import ModelIdVersion
from legion.k8s.definitions import LEGION_COMPONENT_LABEL, LEGION_SYSTEM_LABEL, LEGION_API_SERVICE_PORT
from legion.k8s.definitions import STATUS_OK, STATUS_WARN, STATUS_FAIL
import legion.k8s.utils
from legion.utils import normalize_name

LOGGER = logging.getLogger(__name__)


class Service:
    """
    Data-structure for describing K8S service
    """

    def __init__(self, k8s_service):
        """
        Build service structure from K8S Service

        :param k8s_service: K8S service
        :type k8s_service: V1Service
        """
        if not k8s_service.metadata.labels:
            raise Exception('Invalid service for introspection: labels are missing')

        self._name = k8s_service.metadata.labels.get(LEGION_COMPONENT_LABEL)
        LOGGER.debug('Analyzing service {}'.format(self._name))

        if not self._name:
            raise Exception('Invalid service for introspection: label {} is missing'.format(LEGION_COMPONENT_LABEL))

        self._k8s_service = k8s_service
        # Get port
        ports = self.k8s_service.spec.ports
        if not ports:
            raise Exception('Invalid service for introspection: {} has no ports'.format(self._name))

        api_ports = [port.port for port in ports if port.name == LEGION_API_SERVICE_PORT]
        if not api_ports:
            raise Exception('Invalid service for introspection: cannot find port {} for service {}'
                            .format(LEGION_API_SERVICE_PORT, self._name))

        self._port = api_ports[0]

        self._ingress_data_loaded = False

        self._public_url = None

    @property
    def k8s_service(self):
        """
        Get K8S service

        :return: V1Service -- K8S service
        """
        return self._k8s_service

    @staticmethod
    def is_legion_service(k8s_service):
        """
        Check that k8s service is legion service

        :param k8s_service: K8S service
        :type k8s_service: V1Service
        :return: bool -- is legion service
        """
        return k8s_service.metadata.labels and all(header in k8s_service.metadata.labels
                                                   for header in [LEGION_COMPONENT_LABEL, LEGION_SYSTEM_LABEL])

    def _load_ingress_data(self):
        """
        Load ingress data (lazy loading)

        :return: None
        """
        if self._ingress_data_loaded:
            return

        self._ingress = get_ingress(self._k8s_service.metadata.namespace, self.name)
        if self._ingress:
            self._public_url = get_ingress_url(self._ingress)

        self._ingress_data_loaded = True

    def reload_ingress_cache(self):
        """
        Reload all cached data about ingress

        :return: None
        """
        self._ingress_data_loaded = False
        self._load_ingress_data()

    @property
    def name(self):
        """
        Get name of service (for example, edi)

        :return: str -- name of service
        """
        return self._name

    @property
    def public_url(self):
        """
        Get public http(s) URL for service (from linked ingress).
        HTTPS would be used if it configured in ingress

        :return: str or None -- public URL of service if service linked to ingress
        """
        self._load_ingress_data()
        return self._public_url

    @property
    def internal_domain(self):
        """
        Get internal domain

        :return: str -- internal domain
        """
        return '{}.{}'.format(self.k8s_service.metadata.name, self.k8s_service.metadata.namespace)

    @property
    def internal_port(self):
        """
        Get internal service port (if present)

        :return: int ot None -- internal port
        """
        return self._port

    @property
    def url(self):
        """
        Get internal URL for service

        :return: str -- internal URL
        """
        return 'http://{}:{}'.format(self.internal_domain, self._port)

    @property
    def url_with_ip(self):
        """
        Get internal URL for service with IP instead of name

        :return: str -- internal URL
        """
        return 'http://{}:{}'.format(self.k8s_service.spec.cluster_ip, self._port)

    @property
    def namespace(self):
        """
        Get service namespace

        :return: str -- namespace
        """
        return self.k8s_service.metadata.namespace

    @property
    def enclave(self):
        """
        Get service enclave

        :return: :py:class:`legion.k8s.enclave.Enclave` -- enclave
        """
        return legion.k8s.enclave.Enclave(self.namespace)

    def __str__(self):
        """
        Get string representation

        :return: str
        """
        return 'Service for {}'.format(self.name)

    __repr__ = __str__


class ModelService(Service):
    """
    Data-structure for describing model service
    """

    def __init__(self, k8s_service):
        """
        Build model service structure from K8S Service

        :param k8s_service: K8S service
        :type k8s_service: V1Service
        """
        super().__init__(k8s_service)
        self._model_id = k8s_service.metadata.labels.get(DOMAIN_MODEL_ID)
        self._model_version = k8s_service.metadata.labels.get(DOMAIN_MODEL_VERSION)

        self._deployment = None
        self._deployment_data_loaded = False

    @staticmethod
    def is_model_service(k8s_service):
        """
        Check that k8s service is legion model service

        :param k8s_service: K8S service
        :type k8s_service: V1Service
        :return: bool -- is legion model service
        """
        return k8s_service.metadata.labels and all(header in k8s_service.metadata.labels
                                                   for header in [DOMAIN_MODEL_ID, DOMAIN_MODEL_VERSION])

    @property
    def id(self):
        """
        Get model id

        :return: str -- ID of model
        """
        return self._model_id

    @property
    def version(self):
        """
        Get model version

        :return: str -- version of model
        """
        return self._model_version

    @property
    def id_and_version(self):
        """
        Get ModelIdVersion structure for current model

        :return: :py:class:`legion.k8s.ModelIdVersion`
        """
        return ModelIdVersion(self.id, self.version)

    @property
    def deployment(self):
        """
        Get model deployment

        :return: :py:class:`kubernetes.client.models.extensions_v1beta1_deployment.ExtensionsV1beta1Deployment`
        """
        self._load_deployment_data()
        return self._deployment

    def _load_deployment_data(self):
        """
        Load deployment data (lazy loading)

        :return: None
        """
        if self._deployment_data_loaded:
            return

        client = legion.k8s.utils.build_client()

        extension_api = kubernetes.client.ExtensionsV1beta1Api(client)
        all_deployments = extension_api.list_namespaced_deployment(self._k8s_service.metadata.namespace)
        model_deployments = [deployment for deployment in all_deployments.items
                             if deployment.metadata.labels.get(DOMAIN_MODEL_ID) == self.id
                             and deployment.metadata.labels.get(DOMAIN_MODEL_VERSION) == self.version]

        self._deployment = model_deployments[0] if model_deployments else None

        self._deployment_data_loaded = True

    def reload_cache(self):
        """
        Reload all cached data about ingress

        :return: None
        """
        self._deployment_data_loaded = False
        self._load_deployment_data()

    @property
    def scale(self):
        """
        Ger current scale

        :return: int -- current model scale
        """
        self._load_deployment_data()
        return self.deployment.status.available_replicas if self.deployment.status.available_replicas else 0

    @scale.setter
    def scale(self, new_scale):
        """
        Scale existed model deployment

        :param new_scale: new scale
        :type new_scale: int
        :return: None
        """
        if new_scale < 1:
            raise Exception('Invalid scale parameter: should be greater then 0')

        self._load_deployment_data()
        client = legion.k8s.utils.build_client()

        extension_api = kubernetes.client.ExtensionsV1beta1Api(client)

        old_scale = self._deployment.spec.replicas
        self._deployment.spec.replicas = new_scale

        LOGGER.info('Scaling service {} in namespace {} from {} to {} replicas'
                    .format(self._deployment.metadata.name, self._deployment.metadata.namespace, old_scale, new_scale))

        extension_api.patch_namespaced_deployment(self._deployment.metadata.name,
                                                  self._deployment.metadata.namespace,
                                                  self._deployment)

        self.reload_cache()

    def delete(self, grace_period_seconds=0):
        """
        Remove model from cluster

        :param grace_period_seconds: grace period in seconds
        :type grace_period_seconds: int
        :return: None
        """
        client = legion.k8s.utils.build_client()

        api_instance = kubernetes.client.AppsV1beta1Api(client)
        core_v1api = kubernetes.client.CoreV1Api(client)

        body = kubernetes.client.V1DeleteOptions(propagation_policy='Background',
                                                 grace_period_seconds=grace_period_seconds)

        LOGGER.info('Deleting service {} in namespace {}'.format(self.k8s_service.metadata.name, self.namespace))
        core_v1api.delete_namespaced_service(name=self.k8s_service.metadata.name,
                                             namespace=self.namespace)

        LOGGER.info('Deleting deployment {} in namespace {} with grace period {}s'
                    .format(self.deployment.metadata.name, self.namespace, grace_period_seconds))
        api_instance.delete_namespaced_deployment(name=self.deployment.metadata.name,
                                                  namespace=self.namespace, body=body,
                                                  grace_period_seconds=grace_period_seconds,
                                                  propagation_policy='Background')

    @property
    def desired_scale(self):
        """
        Get desired model scale

        :return: int -- desired model scale
        """
        self._load_deployment_data()
        return self.deployment.status.replicas if self.deployment.status.replicas else 0

    @property
    def status(self):
        """
        Get model status

        :return: str -- model status
        """
        self._load_deployment_data()
        status = STATUS_OK

        if self.scale == 0:
            status = STATUS_FAIL
        elif self.desired_scale > self.scale > 0:
            status = STATUS_WARN

        return status

    @property
    def image(self):
        """
        Get model image

        :return: str -- model image
        """
        self._load_deployment_data()
        return self.deployment.spec.template.spec.containers[0].image

    @property
    def metrics_name(self):
        """
        Get model metrics name

        :return: str -- metrics name
        """
        return '{}.{}'.format(legion.k8s.utils.normalize_name(self.id, dns_1035=True),
                              legion.k8s.utils.normalize_name(self.version, dns_1035=True))

    def __str__(self):
        """
        Get string representation

        :return: str
        """
        return 'ModelService for model {} version {}'.format(self.id, self.version)

    __repr__ = name = __str__


class ModelServiceEndpoint:
    """
    Structure for describing model service with their endpoint (for specific version or default)
    """

    def __init__(self, model_service, default=False):
        """
        Build structure for model service (may used with default flag)

        :param model_service: model service
        :type model_service: :py:class:`legion.k8s.services.ModelService`
        :param default: (Optional) build default (not version locked url)
        :type default: bool
        """
        self._model_service = model_service
        self._default = default

    @property
    def model_service(self):
        """
        Get linked model service

        :return: :py:class:`legion.k8s.services.ModelService` -- model service
        """
        return self._model_service

    @property
    def url(self):
        """
        Get model url

        :return: str -- model url (relative to model API root)
        """
        if self._default:
            return '{}'.format(self.model_service.id)
        else:
            return '{}/{}'.format(self.model_service.id, self.model_service.version)

    def build_default(self):
        """
        Get ModelServiceEndpoint with enabled default flag for current model service

        :return: :py:class:`legion.k8s.services.ModelServiceEndpoint` -- new structure
        """
        return ModelServiceEndpoint(self._model_service, default=True)

    def __str__(self):
        """
        Get string representation

        :return: str
        """
        return 'ModelServiceEndpoint for MS {} default {}, url {}'.format(self.model_service, self._default, self.url)

    __repr__ = __str__

    def __hash__(self):
        """
        Get hash of current endpoint

        :return: int -- hash of current object
        """
        return hash(self.url)

    def __eq__(self, other):
        """
        Check equality

        :return: bool -- is items equal
        """
        return self.url == other.url


def find_model_deployment(model_id, namespace='default'):
    """
    Find model deployment by model id

    :param model_id: model id
    :type model_id: str
    :param namespace: namespace
    :type namespace: str
    :return: :py:class:`kubernetes.client.models.extensions_v1beta1_deployment.ExtensionsV1beta1Deployment`
    """
    client = legion.k8s.utils.build_client()

    extension_api = kubernetes.client.ExtensionsV1beta1Api(client)
    all_deployments = extension_api.list_namespaced_deployment(namespace)

    type_label_name = normalize_name(legion.containers.headers.DOMAIN_CONTAINER_TYPE)
    type_label_value = 'model'

    model_id_name = normalize_name(legion.containers.headers.DOMAIN_MODEL_ID)
    model_id_value = normalize_name(model_id)

    for deployment in all_deployments.items:
        if deployment.metadata.labels.get(type_label_name) == type_label_value \
                and deployment.metadata.labels.get(model_id_name) == model_id_value:
            return deployment

    return None


def find_all_models_deployments(namespace='default'):
    """
    Find all models deployments

    :param namespace: namespace
    :type namespace: str or none for all
    :return: list[:py:class:`kubernetes.client.models.extensions_v1beta1_deployment.ExtensionsV1beta1Deployment`]
    """
    client = legion.k8s.utils.build_client()

    extension_api = kubernetes.client.ExtensionsV1beta1Api(client)
    if namespace:
        all_deployments = extension_api.list_namespaced_deployment(namespace)
    else:
        all_deployments = extension_api.list_deployment_for_all_namespaces()

    type_label_name = normalize_name(legion.containers.headers.DOMAIN_CONTAINER_TYPE)
    type_label_value = 'model'

    model_deployments = [
        deployment
        for deployment in all_deployments.items
        if deployment.metadata.labels.get(type_label_name) == type_label_value
    ]

    return model_deployments


def find_all_services(namespace='', component=''):
    """
    Find all services details by criteria

    :param namespace: namespace
    :type namespace: str or none for all
    :param component: filter by specified component value, or none for all
    :type component: str
    :return: list[V1Service]
    """
    client = legion.k8s.utils.build_client()

    core_api = kubernetes.client.CoreV1Api(client)
    if namespace:
        all_services = core_api.list_namespaced_service(namespace)
    else:
        all_services = core_api.list_service_for_all_namespaces()

    if component:
        all_services = [
            service
            for service in all_services.items
            if service.metadata.labels.get(LEGION_COMPONENT_LABEL) == component
        ]
    else:
        all_services = all_services.items

    return all_services


def get_service(namespace='', component=''):
    """
    Get a service details by criteria

    :param namespace: namespace
    :type namespace: str or none
    :param component: filter by specified component value, or none for all
    :type component: str
    :return: V1Service or None
    """
    result = find_all_services(namespace, component)
    return legion.k8s.services.Service(result[0]) if result else None


def find_all_ingresses(namespace='', component=''):
    """
    Find all ingresses details by criteria

    :param namespace: namespace
    :type namespace: str or none for all
    :type component: filter by specified component value, or none for all
    :return: list[V1beta1Ingress]
    """
    client = legion.k8s.utils.build_client()

    extension_api = kubernetes.client.ExtensionsV1beta1Api(client)
    if namespace:
        all_ingresses = extension_api.list_namespaced_ingress(namespace)
    else:
        all_ingresses = extension_api.list_ingress_for_all_namespaces()

    if component:
        all_ingresses = [
            ingress
            for ingress in all_ingresses.items
            if ingress.metadata.labels.get(LEGION_COMPONENT_LABEL) == component
        ]
    else:
        all_ingresses = all_ingresses.items

    return all_ingresses


def get_ingress(namespace='', component=''):
    """
    Get ingress details by criteria

    :param namespace: namespace
    :type namespace: str or none for all
    :type component: filter by specified component value, or none for all
    :return: V1beta1Ingress -- Ingress object or None
    """
    ingresses = find_all_ingresses(namespace, component)
    return ingresses[0] if ingresses else None


def get_ingress_url(ingress):
    """
    Get URL of ingress object

    :param ingress: ingress object
    :type ingress: V1beta1Ingress
    :return: str or None -- URL from tls section,
    """
    spec = ingress.spec

    host = spec.rules[0].host if spec.rules else None
    if not host:
        return None

    protocol = 'https' if spec.tls and any(host in tls.hosts for tls in spec.tls) else 'http'

    return '{}://{}'.format(protocol, host)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Flask package
"""
import functools
import os
import logging
import urllib

import legion.config
import legion.utils

import flask
from requests.compat import urlencode
from requests.utils import to_key_val_list
from urllib.parse import parse_qs

LOGGER = logging.getLogger(__name__)


def encode_http_params(data):
    """
    Encode HTTP parameters to URL query string

    :param data: data as text or tuple/list
    :type data: str or bytes or tuple or list
    :return: str -- encoded data
    """
    if isinstance(data, (str, bytes)):
        return urlencode(data)
    elif hasattr(data, '__iter__'):
        result = []
        for k, vs in to_key_val_list(data):
            if vs is not None:
                result.append(
                    (k.encode('utf-8') if isinstance(k, str) else k,
                     vs.encode('utf-8') if isinstance(vs, str) else vs))
        return urlencode(result, doseq=True)
    else:
        raise ValueError('Invalid argument')


def parse_multi_dict(multi_dict, map=None):
    """
    Parse MultiDict object and detect lists if they presents (using [] in the end of name)

    :param multi_dict: request data
    :type multi_dict: :py:class:`werkzeug.datastructures.MultiDict`
    :param map: function to map all values
    :type map: function VAL -> NEW VAL
    :return: dict[str, Any] or dict[str, list[Any]]
    """
    result = {}

    for k in multi_dict:
        if len(k) > 2 and k.endswith('[]'):
            key = k[:-2]
            result[key] = multi_dict.getlist(k)
            if map:
                result[key] = [map(val) for val in result[key]]
        else:
            result[k] = multi_dict[k]
            if map:
                result[k] = map(result[k])

    return result


def parse_url_querystring(querystring_dict):
    """
    Parse URL query strings dictionaries like {'a': ['123'], 'b[]': ['one', 'two']}
    to appropriate dictionaries ({'a': '123', 'b': ['one', 'two']})

    :param querystring_dict: querystring dictionary
    :type querystring_dict: dict
    :return: dict -- parsed and flatted querystring dictionary
    """
    result = {}
    for k in querystring_dict:
        if k.endswith('[]'):
            key = k[:-2]
            result[key] = querystring_dict[k]
        else:
            result[k] = querystring_dict[k][0]
    return result


def parse_batch_request(input_request):
    """
    Parse request in batch mode with payload encoded in body

    :param input_request: request object
    :type input_request: :py:class:`Flask.request`
    :return: list[dict] -- list of dicts with requested fields
    """
    if not input_request.data:
        raise Exception('Request does not contain any data')

    return [parse_url_querystring(parse_qs(line))
            for line in input_request.data.decode('utf-8').split('\n')]


def parse_request(input_request):
    """
    Produce a input dictionary from HTTP request (GET/POST fields, and Files)

    :param input_request: request object
    :type input_request: :py:class:`Flask.request`
    :return: dict with requested fields
    """
    result = {}

    # Fill in URL parameters
    result.update(parse_multi_dict(input_request.args))

    # Fill in POST parameters
    result.update(parse_multi_dict(input_request.form))

    # Fill in Files:
    result.update(parse_multi_dict(input_request.files, lambda file: file.read()))

    return result


def prepare_response(response):
    """
    Produce an HTTP response from dict/list

    :param response: dict/list with data
    :type response: dict[str, any] or list[any]
    :return: bytes
    """
    return flask.jsonify(response)


def provide_json_response(method):
    """
    Process response data via JSON formatter

    :param method: decorator function
    :return: decorated function
    """
    @functools.wraps(method)
    def decorated_function(*args, **kwargs):
        code = 200
        try:
            response = method(*args, **kwargs)
            if isinstance(response, bool):
                response = {'status': response}
            elif not isinstance(response, dict) and not isinstance(response, list):
                raise Exception('Wrong value returned from API handler')
        except Exception as exception:
            code = 500
            response = {'error': True,
                        'exception': str(exception)}
            LOGGER.exception('Exception during processing request for {}: {}'
                             .format(method.__name__, exception),
                             exc_info=exception)

        response = prepare_response(response)
        return flask.make_response(response, code)

    return decorated_function


def populate_fields(**fields):
    """
    Populate fields data

    :param fields: fields with formats
    :type fields: dict[str, type]
    :return: decorator function
    """
    def decorator(method):
        @functools.wraps(method)
        def decorated_function(*args, **kwargs):
            parameters = parse_request(flask.request)
            casted_parameters = {}

            for field, value in parameters.items():
                if field not in fields:
                    raise Exception('Unknown parameter: %s' % field)

                try:
                    casted_parameters[field] = fields[field](value)
                except ValueError as cast_value_exception:
                    raise Exception('Cannot cast field %s to %s: %s' % (field, fields[field], cast_value_exception))

            kwargs.update(casted_parameters)

            return method(*args, **kwargs)

        return decorated_function

    return decorator


def authenticate(authenticator):
    """
    Authenticate user

    :param authenticator: callable obj. for checking credentials. Arguments are user and password. Should returns bool.
    :type authenticator: function[user, password]: bool
    :return: decorator function
    """
    def decorator(method):
        @functools.wraps(method)
        def decorated_function(*args, **kwargs):
            auth = flask.request.authorization
            username = None
            password = None

            if auth:
                username = auth.username
                password = auth.password

            if not authenticator(username, password):
                raise legion.utils.EdiHTTPAccessDeniedException()

            return method(*args, **kwargs)

        return decorated_function

    return decorator


def requested_fields(*fields):
    """
    Check that all requested fields has been passed

    :param fields: list of field names
    :type fields: list[str]
    :return: decorator function
    """
    def decorator(method):
        @functools.wraps(method)
        def decorated_function(*args, **kwargs):
            for field in fields:
                if field not in kwargs:
                    raise Exception('Requested field %s s not set' % field)

            return method(*args, **kwargs)

        return decorated_function

    return decorator


def apply_cli_args(application, args):
    """
    Set Flask app instance configuration from arguments

    :param application: Flask app instance
    :type application: :py:class:`Flask.app`
    :param args: arguments
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    args_dict = vars(args)
    for argument, value in args_dict.items():
        if value is not None:
            application.config[argument.upper()] = value


def apply_env_argument(application, name, cast=None):
    """
    Update application config if ENV variable exists

    :param application: Flask app instance
    :type application: :py:class:`Flask.app`
    :param name: environment variable name
    :type name: str
    :param cast: casting of str variable
    :type cast: Callable[[str], Any]
    :return: None
    """
    if name in os.environ:
        value = os.getenv(name)
        if cast:
            value = cast(value)

        application.config[name] = value


def apply_env_args(application):
    """
    Set Flask app instance configuration from environment

    :param application: Flask app instance
    :type application: :py:class:`Flask.app`
    :return: None
    """
    apply_env_argument(application, legion.config.MODEL_ID[0])
    apply_env_argument(application, legion.config.MODEL_FILE[0])

    apply_env_argument(application, legion.config.LEGION_ADDR[0])
    apply_env_argument(application, legion.config.LEGION_PORT[0], cast=int)

    apply_env_argument(application, legion.config.DEBUG[0], legion.utils.string_to_bool)

    apply_env_argument(application, legion.config.REGISTER_ON_GRAFANA[0], legion.utils.string_to_bool)

    apply_env_argument(application, legion.config.NAMESPACE[0])

    apply_env_argument(application, legion.config.LEGION_API_ADDR[0])
    apply_env_argument(application, legion.config.LEGION_API_PORT[0], cast=int)

    apply_env_argument(application, legion.config.CLUSTER_CONFIG_PATH[0])
    apply_env_argument(application, legion.config.CLUSTER_SECRETS_PATH[0])


def configure_application(application, args):
    """
    Initialize configured Flask application instance
    Overall configuration priority: config_default.py, env::FLASK_APP_SETTINGS_FILES file,
    ENV parameters, CLI parameters

    :param application: Flask app instance
    :type application: :py:class:`Flask.app`
    :param args: arguments if provided
    :type args: :py:class:`argparse.Namespace` or None
    :return: None
    """
    # 4th priority: config from file with defaults values
    application.config.from_pyfile('config_default.py')

    # 3rd priority: config from file (path to file from ENV)
    application.config.from_envvar(legion.config.FLASK_APP_SETTINGS_FILES[0], True)

    # 2nd priority: config from ENV variables
    apply_env_args(application)

    # 1st priority: config from CLI args
    if args:
        apply_cli_args(application, args)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s watch context manager
"""
import logging

import kubernetes
import kubernetes.client
import kubernetes.config
import kubernetes.config.config_exception
import urllib3
import urllib3.exceptions

LOGGER = logging.getLogger(__name__)
CONNECTION_CONTEXT = None


class ResourceWatch:
    """
    Watch for K8S resources (context manager)
    """

    def __init__(self, api_function, *args,
                 filter_callable=None,
                 object_constructor=None,
                 **kwargs):
        """
        Initialize context manager for resource watch

        :param api_function: API function to watch
        :param args: additional positional arguments for API function
        :param filter_callable: (Optional) callable to filter objects
        :param object_constructor:  (Optional) callable to construct object wrappers
        :param kwargs: additional key value arguments for API function
        """
        self._api_function = api_function
        self._filter_callable = filter_callable
        self._object_constructor = object_constructor
        self._args = args
        self._kwargs = kwargs

        self._watch = kubernetes.watch.Watch()
        self._stream = None

    def __enter__(self):
        """
        Enter watch

        :return:
        """
        self._stream = self._watch.stream(self._api_function,
                                          *self._args,
                                          **self._kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit watch

        :param exc_type: exception type
        :param exc_val: exception
        :param exc_tb: exception traceback
        :return: None
        """
        self.stop()

    @property
    def stream(self):
        """
        Access watch stream

        :return: tuple(str, Any) -- event type and event object (plain or constructed if
                                    object_constructor has been passed)
        """
        LOGGER.debug('Starting watch stream')

        reconnect = True
        while reconnect:
            reconnect = False

            try:
                for event in self._stream:
                    event_type = event['type']
                    event_object = event['object']

                    # Check if valid object
                    pass_event = not self._filter_callable or self._filter_callable(event_object)

                    if pass_event:
                        # Construct specific object
                        if self._object_constructor:
                            event_object = self._object_constructor(event_object)

                        yield (event_type, event_object)
            except urllib3.exceptions.ProtocolError:
                LOGGER.info('Connection to K8S API has been lost, reconnecting..')
                reconnect = True

        LOGGER.debug('Watch stream has been ended')

    @property
    def watch(self):
        """
        Access watch object

        :return: :py:class:`kubernetes.watch.Watch` -- watch object
        """
        return self._watch

    def stop(self):
        """
        Stop watch

        :return: None
        """
        try:
            self.watch.stop()
        finally:
            LOGGER.debug('Watch has been stopped')



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
EDI client
"""

import json
import logging
import os

import legion.k8s
import legion.edi.server
import legion.config
import requests
import requests.exceptions

LOGGER = logging.getLogger(__name__)


class EdiClient:
    """
    EDI client
    """

    def __init__(self, base, user=None, password=None, token=None):
        """
        Build client

        :param base: base url, for example: http://edi.parallels
        :type base: str
        :param user: user name for user/password based auth
        :type user: str or None
        :param password: user password for user/password based auth
        :type password: str or None
        :param token: token for token based auth
        :type token: str or None
        """
        self._base = base
        self._user = user
        self._password = password
        self._token = token
        self._version = legion.edi.server.EDI_VERSION

    def _query(self, url_template, payload=None, action='GET'):
        """
        Perform query to EDI server

        :param url_template: url template from legion.const.api
        :type url_template: str
        :param payload: payload (will be converted to JSON) or None
        :type payload: dict[str, any]
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :type action: str
        :return: dict[str, any] -- response content
        """
        sub_url = url_template.format(version=self._version)
        full_url = self._base.strip('/') + sub_url

        auth = None
        headers = {}

        if self._user and self._password:
            auth = (self._user, self._password)
        elif self._token:
            auth = ('token', self._token)

        try:
            response = requests.request(action.lower(), full_url, data=payload, headers=headers, auth=auth)
        except requests.exceptions.ConnectionError as exception:
            raise Exception('Failed to connect to {}: {}'.format(self._base, exception))

        LOGGER.debug('Got answer: {!r} with code {} for URL {!r}'
                     .format(response.text, response.status_code, full_url))

        try:
            answer = json.loads(response.text)
        except ValueError as json_decode_exception:
            raise ValueError('Invalid JSON structure {!r}: {}'.format(response.text, json_decode_exception))

        if isinstance(answer, dict) and answer.get('error', False):
            exception = answer.get('exception')
            raise Exception('Got error from server: {!r}'.format(exception))

        if response.status_code != 200:
            raise Exception('Server returned wrong HTTP code (not 200) without error flag')

        return answer

    def inspect(self, model=None, version=None):
        """
        Perform inspect query on EDI server

        :param model: model id
        :type model: str
        :param version: (Optional) model version
        :type version: str
        :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
        """
        payload = {}
        if model:
            payload['model'] = model
        if version:
            payload['version'] = version

        answer = self._query(legion.edi.server.EDI_INSPECT, payload=payload)
        return [legion.k8s.ModelDeploymentDescription(**x) for x in answer]

    def info(self):
        """
        Perform info query on EDI server

        :return: dict[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
        """
        return self._query(legion.edi.server.EDI_INFO)

    def deploy(self, image, count=1):
        """
        Deploy API endpoint

        :param image: Docker image for deploy (for kubernetes deployment and local pull)
        :type image: str
        :param count: count of pods to create
        :type count: int
        :return: bool -- True
        """
        payload = {
            'image': image
        }
        if count is not None:
            payload['count'] = count

        return self._query(legion.edi.server.EDI_DEPLOY, action='POST', payload=payload)['status']

    def undeploy(self, model, grace_period=0, version=None):
        """
        Undeploy API endpoint

        :param model: model id
        :type model: str
        :param grace_period: grace period for removing
        :type grace_period: int
        :param version: (Optional) model version
        :type version: str
        :return: bool -- True
        """
        payload = {
            'model': model
        }

        if grace_period:
            payload['grace_period'] = grace_period

        if version:
            payload['version'] = version

        return self._query(legion.edi.server.EDI_UNDEPLOY, action='POST', payload=payload)['status']

    def scale(self, model, count, version=None):
        """
        Scale model

        :param model: model id
        :type model: str
        :param count: count of pods to create
        :type count: int
        :param version: (Optional) model version
        :type version: str
        :return: bool -- True
        """
        payload = {
            'model': model,
            'count': count
        }
        if version:
            payload['version'] = version

        return self._query(legion.edi.server.EDI_SCALE, action='POST', payload=payload)['status']


def add_edi_arguments(parser):
    """
    Add EDI arguments parser

    :param parser:
    :type parser:
    :return:
    """
    parser.add_argument('--edi',
                        type=str, help='EDI server host')
    parser.add_argument('--user',
                        type=str, help='EDI server user')
    parser.add_argument('--password',
                        type=str, help='EDI server password')
    parser.add_argument('--token',
                        type=str, help='EDI server token')


def build_client(args):
    """
    Build EDI client from from ENV and from command line arguments

    :param args: command arguments with .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: :py:class:`legion.external.edi.EdiClient` -- EDI client
    """
    host = os.environ.get(*legion.config.EDI_URL)
    user = os.environ.get(*legion.config.EDI_USER)
    password = os.environ.get(*legion.config.EDI_PASSWORD)
    token = os.environ.get(*legion.config.EDI_TOKEN)

    if args.edi:
        host = args.edi

    if args.user:
        user = args.user

    if args.password:
        password = args.password

    if args.token:
        token = args.token

    client = EdiClient(host, user, password, token)
    return client



#
#   Copyright 2018 EPAM Systems
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
Module fore remote debugging python code in docker containers
PYDEVD_HOST and PYDEVD_PORT should be set to valid values
"""
import os
import logging

LOGGER = logging.getLogger(__name__)

try:
    host = os.getenv('PYDEVD_HOST')
    port = os.getenv('PYDEVD_PORT')
    LOGGER.debug('Docker bootup got configuration: host={!r} port={!r}'.format(host, port))

    if host and port:
        import pydevd

        try:
            port = int(port)
        except ValueError as value_error:
            LOGGER.exception('Invalid port value: {}'.format(port), exc_info=value_error)

        pydevd.settrace(
            host=host,
            port=int(port),
            suspend=False,
            patch_multiprocessing=True,
            stdoutToServer=True,
            stderrToServer=True)
except Exception as pydevd_connection_exception:
    LOGGER.exception('Unable to connect to pydevd. Ignoring...', exc_info=pydevd_connection_exception)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Enclave models Notifier module for Legion Template
"""
import logging

from legion.k8s import Enclave
from legion.k8s.utils import get_current_namespace

LOGGER = logging.getLogger(__name__)


async def enclave_models_monitor(template_system):
    """
    Update template with current model state

    :param template_system: Object, that contains 'render' callback function
    :return: None
    """
    namespace = get_current_namespace()
    LOGGER.info('Starting models monitor in namespace {}'.format(namespace))

    enclave = Enclave(namespace)
    LOGGER.info('Loaded enclave {}'.format(enclave))

    for new_state in enclave.watch_model_service_endpoints_state():
        LOGGER.info('Got updated model state')
        template_system.render(models=new_state)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s utils functions
"""
import os
import os.path
import logging

import docker
import docker.errors
import kubernetes
import kubernetes.client
import kubernetes.config
import kubernetes.config.config_exception
import urllib3
import urllib3.exceptions
import yaml

import legion
import legion.containers.docker
import legion.containers.headers
import legion.config
import legion.external.grafana
from legion.utils import normalize_name
import legion.k8s.services
import legion.k8s.enclave
from legion.k8s.definitions import \
    LEGION_COMPONENT_LABEL, LEGION_COMPONENT_NAME_MODEL, \
    LEGION_SYSTEM_LABEL, LEGION_SYSTEM_VALUE

LOGGER = logging.getLogger(__name__)
CONNECTION_CONTEXT = None
KUBERNETES_SERVICE_ACCOUNT_NAMESPACE_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'


def build_client():
    """
    Configure and returns kubernetes client

    :return: :py:module:`kubernetes.client`
    """
    try:
        kubernetes.config.load_incluster_config()
    except kubernetes.config.config_exception.ConfigException:
        kubernetes.config.load_kube_config(context=CONNECTION_CONTEXT)

    # Disable SSL warning for self-signed certificates
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    return kubernetes.client.ApiClient()


def get_current_namespace():
    """
    Get current namespace [WORKS ONLY IN-CLUSTER]

    :return: str -- name of namespace
    """
    try:
        with open(KUBERNETES_SERVICE_ACCOUNT_NAMESPACE_PATH, 'r') as namespace_file:
            return namespace_file.read()
    except Exception as exception:
        raise Exception('Cannot get current namespace (it works only in cluster): {}'.format(exception))


def is_code_run_in_cluster():
    """
    Is current code run in a cluster

    :return: bool -- is code run in a cluster or not
    """
    return os.path.exists(KUBERNETES_SERVICE_ACCOUNT_NAMESPACE_PATH)


def load_config(path_to_config):
    """
    Load cluster config

    :param path_to_config: path to cluster config file
    :type path_to_config: str
    :return: dict -- cluster config
    """
    if not os.path.exists(path_to_config):
        raise Exception('config path %s not exists' % path_to_config)

    if not os.path.isfile(path_to_config):
        raise Exception('config path %s is not a file' % path_to_config)

    with open(path_to_config, 'r') as stream:
        return yaml.load(stream)


def load_secrets(path_to_secrets):
    """
    Load all secrets from folder to dict

    :param path_to_secrets: path to secrets directory
    :type path_to_secrets: str
    :return: dict[str, str] -- dict of secret name => secret value
    """
    if not os.path.exists(path_to_secrets):
        raise Exception('secrets path %s not exists' % path_to_secrets)

    if not os.path.isdir(path_to_secrets):
        raise Exception('secrets path %s is not a directory' % path_to_secrets)

    files = [
        (file, os.path.abspath(os.path.join(path_to_secrets, file))) for file
        in os.listdir(path_to_secrets)
        if os.path.isfile(os.path.join(path_to_secrets, file))
    ]

    secrets = {}

    for name, path in files:
        try:
            with open(path, 'r') as stream:
                secrets[name] = stream.read()
                if isinstance(secrets[name], bytes):
                    secrets[name] = secrets[name].decode('utf-8')
        except IOError:
            pass
    return secrets


def get_docker_image_labels(image):
    """
    Get labels from docker image

    :param image: docker image
    :type image: str
    :return: dict[str, Any] -- image labels
    """
    docker_client = legion.containers.docker.build_docker_client(None)
    try:
        try:
            docker_image = docker_client.images.get(image)
        except docker.errors.ImageNotFound:
            docker_image = docker_client.images.pull(image)
    except Exception as docker_pull_exception:
        raise Exception('Cannot pull docker image {}: {}'.format(image, docker_pull_exception))

    required_headers = [
        legion.containers.headers.DOMAIN_MODEL_ID,
        legion.containers.headers.DOMAIN_MODEL_VERSION,
        legion.containers.headers.DOMAIN_CONTAINER_TYPE
    ]

    if any(header not in docker_image.labels for header in required_headers):
        raise Exception('Missed one of %s labels. Available labels: %s' % (
            ', '.join(required_headers),
            ', '.join(tuple(docker_image.labels.keys()))
        ))

    return docker_image.labels


def get_meta_from_docker_labels(labels):
    """
    Build meta fields for kubernetes from docker labels.

    :param labels: docker image labels
    :type labels: dict[str, Any]
    :return: tuple[str, dict[str, str], str, str] -- k8s_name name, labels in DNS-1123 format, model id and version
    """
    model_id = labels[legion.containers.headers.DOMAIN_MODEL_ID]
    model_version = labels[legion.containers.headers.DOMAIN_MODEL_VERSION]

    k8s_name = "model-%s-%s" % (
        normalize_name(model_id),
        normalize_name(model_version)
    )

    compatible_labels = {
        normalize_name(k, kubernetes_compatible=True): normalize_name(v, kubernetes_compatible=True)
        for k, v in
        labels.items()
        if k.startswith(legion.containers.headers.DOMAIN_PREFIX)
    }

    compatible_labels[LEGION_COMPONENT_LABEL] = LEGION_COMPONENT_NAME_MODEL
    compatible_labels[LEGION_SYSTEM_LABEL] = LEGION_SYSTEM_VALUE

    return normalize_name(k8s_name, dns_1035=True), compatible_labels, model_id, model_version



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

"""
Template system that uses courutines
"""
import logging
import os
import os.path
import importlib

import asyncio
from asyncio.tasks import Task

from jinja2 import Environment, FileSystemLoader, Undefined

LOGGER = logging.getLogger(__name__)


class LegionTemplateEngine:
    """
    Extensible template system
    """

    def __init__(self, template_file, output_file, command=None, signal=None, pid=None, pid_file=None):
        """
        Initialize template system

        :param template_file: source file
        :type template_file: str
        :param output_file: target file
        :type output_file: str
        :param command: (Optional) bash command to execute on file render
        :type command: str
        :param signal: (Optional) signal id, that is sent to a process on render
        :type signal: int
        :param pid: (Optional) process id, that receives signal on render
        :type pid: int
        :param pid_file: (Optional) process id file, that receives signal on render
        :type pid_file: str
        """
        LOGGER.debug('Creating LegionTemplateEngine instance for source={}, destination={}'
                     .format(template_file, output_file))
        self._template_file = template_file
        self._output_file = output_file
        self._command = command
        self._signal = signal
        self._pid = pid
        self._pid_file = pid_file

        if signal and not pid and not pid_file:
            raise ValueError('PID or PID file parameters is required, if signal is specified')

        self._j2 = Environment(
            loader=FileSystemLoader([os.getcwd(), '/']),
            undefined=SilentUndefined
        )
        self._template = self._j2.get_template(template_file)

        self._context = {
            'load_module': self.load_module
        }
        self._coroutines = []
        self._loop = asyncio.get_event_loop()
        self._initializing_mode = True

    @property
    def template_file_path(self):
        """
        Get path to template file

        :return: str -- path to template file
        """
        return self._template_file

    @property
    def output_file_path(self):
        """
        Get output path

        :return: str -- output path
        """
        return self._output_file

    def load_module(self, name, *args, **kwargs):
        """
        Load module
        Example: {{ load_module('loadable_module', *args, *kwargs) }}
        and module should be
            async def temp(template_system, *args, **kwargs):
                print("I'm in template extension")
                for i in range(2):
                    template_system.render(a=i)
                    await asyncio.sleep(1)

        :param name: name of module
        :type name: str
        :param args: additional positional arguments
        :type args: any
        :param kwargs: additional key-value arguments
        :type args: any
        :return: str -- empty string
        """
        if not self._initializing_mode:
            return ''

        LOGGER.info('Loading module {} with arguments {} and {}'.format(name, args, kwargs))

        module_name, function_name = name.rsplit('.', 1)
        module = importlib.import_module(module_name)
        module_function = getattr(module, function_name)
        future = asyncio.ensure_future(module_function(self, *args, **kwargs))

        self._coroutines.append(future)
        return ''

    def render(self, **items):
        """
        Render template and notify targets (execute bash command or send signal)

        :param items: values to update context
        :type items: dict[str, Any]
        :return: None
        """
        LOGGER.debug('Updating context')
        self._context.update(items)

        # Interrupt rendering if we are in initializing mode
        if self._initializing_mode:
            return None

        LOGGER.debug('Rendering')
        # Render and update target file
        content = self._template.render(self._context)
        with open(self._output_file, 'w') as file_stream:
            file_stream.write(content)

        # Notify targets
        if self._command:
            os.system(self._command)
        if self._signal and self._pid:
            os.kill(self._pid, self._signal)
        if self._signal and self._pid_file and os.path.exists(self._pid_file):
            with open(self._pid_file) as pf:
                pid = pf.read().strip()
                if not pid.isdigit():
                    raise ValueError('PID "{}" is not an integer'.format(pid))
                os.kill(int(pid), self._signal)

    def render_loop(self):
        """
        Start render loop

        :return: None
        """
        # Load modules by rendering template
        LOGGER.debug('Staring module initialization')
        self._initializing_mode = True
        self._template.render(self._context)
        self._initializing_mode = False
        LOGGER.debug('Modules have been initialized')

        # Do first render
        LOGGER.debug('Staring first render')
        self.render()

        # Check is any courutine present
        if not self._coroutines:
            raise Exception('Template doesnt use any plugin')

        # Start loop
        LOGGER.debug('Staring loop')
        result = self._loop.run_until_complete(asyncio.wait(self._coroutines))
        LOGGER.debug('Loop finished')

        if result is not None and type(result) == tuple and len(result) > 0 and type(result[0]) == set:
            task = result[0].pop()
            if type(task) == Task and task.exception() is not None:
                raise task.exception()


class SilentUndefined(Undefined):
    """
    SilentUndefined class is a replacement for original Undefined class in Jinja2,
    which is used to raise an Error if undefined variable is met in template.
    Now it doesn't raise an error, just writes empty string.
    """

    def _fail_with_undefined_error(self, *args, **kwargs):
        """
        Return an empty string instead of raise an Error.
        It's a replacement for an original _fail_with_undefined_error
        function, which is used to raise an Error.
        """
        class EmptyString(str):
            """
            A simple empty string wrapper.
            """

            def __call__(self, *args, **kwargs):
                """
                Overload call method

                :param args: args
                :param kwargs: kwargs
                :return: str -- empty string
                """
                return ''

        return EmptyString()

    __add__ = __radd__ = __mul__ = __rmul__ = __div__ = __rdiv__ = \
        __truediv__ = __rtruediv__ = __floordiv__ = __rfloordiv__ = \
        __mod__ = __rmod__ = __pos__ = __neg__ = __call__ = \
        __getitem__ = __lt__ = __le__ = __gt__ = __ge__ = __int__ = \
        __float__ = __complex__ = __pow__ = __rpow__ = \
        _fail_with_undefined_error



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion utils functional
"""
import datetime
import contextlib
import os
import getpass
import distutils.dir_util
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import zipfile

import legion.config
import legion.containers.headers

import docker
import requests
import requests.auth
from jinja2 import Environment, PackageLoader, select_autoescape


KUBERNETES_STRING_LENGTH_LIMIT = 63


def render_template(template_name, values=None):
    """
    Render template with parameters
    :param template_name: name of template without path (all templates should be placed in legion.templates directory)
    :param values: dict template variables or None
    :return: str rendered template
    """
    env = Environment(
        loader=PackageLoader(__name__, 'templates'),
        autoescape=select_autoescape(['tmpl'])
    )

    if not values:
        values = {}

    template = env.get_template(template_name)
    return template.render(values)


class EdiHTTPException(Exception):
    """
    Exception for EDI server (with HTTP code)
    """

    def __init__(self, http_code, message):
        """
        Build exception

        :param http_code: HTTP code
        :type http_code: int
        :param message: message for user
        :type message: str
        """
        super(EdiHTTPException, self).__init__(message)
        self.http_code = http_code
        self.message = message


class EdiHTTPAccessDeniedException(EdiHTTPException):
    """
    Exception for EDI server -- access denied
    """

    def __init__(self, message='Access denied'):
        """
        Build exception

        :param message: message for user
        :type message: str
        """
        super(EdiHTTPAccessDeniedException, self).__init__(403, message)


def detect_ip():
    """
    Get current machine IP address

    :return: str -- IP address
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.connect(("8.8.8.8", 80))
    addr = sock.getsockname()[0]
    sock.close()
    return addr


def escape(unescaped_string):
    """
    Escape string (replace .:& with -)

    :param unescaped_string: source string
    :type unescaped_string: str
    :return: str -- escaped string
    """
    return unescaped_string.replace('.', '-').replace(':', '-').replace('&', '-')


def remove_directory(path):
    """
    Remove directory and all subdirectories or file

    :param path: path to directory or file
    :type path: str
    :return: None
    """
    try:
        if os.path.isdir(path):
            for root, dirs, files in os.walk(path, topdown=False):
                for name in files:
                    os.remove(os.path.join(root, name))
                for name in dirs:
                    os.rmdir(os.path.join(root, name))

            os.rmdir(path)
        elif os.path.isfile(path):
            os.remove(path)
        else:
            raise Exception('Not a directory or file: %s' % path)
    finally:
        pass


class TemporaryFolder:
    """
    Temporary folder representation with context manager (temp. directory deletes of context exit)
    """

    def __init__(self, *args, **kwargs):
        """
        Build temp. folder representation using tempfile.mkdtemp

        :param args: tempfile.mkdtemp args
        :type args: tuple
        :param kwargs: tempfile.mkdtemp kwargs
        :type kwargs: dict
        """
        self._path = tempfile.mkdtemp(*args, **kwargs)

    @property
    def path(self):
        """
        Get path to temp. folder

        :return: str -- path
        """
        return self._path

    def remove(self):
        """
        Try to remove temporary folder (without exceptions)

        :return: None
        """
        remove_directory(self._path)

    def __enter__(self):
        """
        Return self on context enter

        :return: :py:class:`legion.utils.TemporaryFolder`
        """
        return self

    def __exit__(self, exit_type, value, traceback):
        """
        Call remove on context exit

        :param exit_type: -
        :param value: -
        :param traceback: -
        :return: None
        """
        self.remove()


class Colors:
    """
    Terminal colors
    """

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def normalize_name(name, dns_1035=False, kubernetes_compatible=False):
    """
    Normalize name

    :param name: name to normalize
    :type name: str
    :param dns_1035: (Optional) use DNS-1035 format, by default False
    :type dns_1035: bool
    :param kubernetes_compatible: (Optional) fit into kubernetes length limitations
    :type kubernetes_compatible: bool
    :return: str -- normalized name
    """
    invalid_delimiters = ' ', '_', '+'
    invalid_chars = '[^a-zA-Z0-9\-\.]'
    if dns_1035:
        invalid_chars = '[^a-zA-Z0-9\-]'
        invalid_delimiters = ' ', '_', '+', '.'

    for char in invalid_delimiters:
        name = name.replace(char, '-')

    value = re.sub(invalid_chars, '', name)
    if kubernetes_compatible:
        value = value[:KUBERNETES_STRING_LENGTH_LIMIT]

    return value


def is_local_resource(path):
    """
    Check if path is local resource

    :param path:
    :type path: str
    :return: bool -- is path for local resource
    """
    if any(path.lower().startswith(prefix) for prefix in ['http://', 'https://', '//']):
        return False

    if '://' in path:
        raise Exception('Unknown or unavailable resource: %s' % path)

    return True


def normalize_external_resource_path(path):
    """
    Normalize external resource path

    :param path: non-normalized resource path
    :type path: str
    :return: str -- normalized path
    """
    default_protocol = os.getenv(*legion.config.EXTERNAL_RESOURCE_PROTOCOL)
    default_host = os.getenv(*legion.config.EXTERNAL_RESOURCE_HOST)

    if path.lower().startswith('//'):
        path = '%s:%s' % (default_protocol, path)

    first_double_slash = path.find('//')

    if first_double_slash < 0:
        raise Exception('Cannot found double slash')

    if len(path) == first_double_slash + 2:
        return path + default_host
    elif len(path) > first_double_slash + 2 and path[first_double_slash + 2] == '/':
        return path[:first_double_slash] + '//' + default_host + path[first_double_slash + 2:]

    return path


def _get_auth_credentials_for_external_resource():
    """
    Get HTTP auth credentials for requests module

    :return: :py:class:`requests.auth.HTTPBasicAuth` -- credentials
    """
    user = os.getenv(*legion.config.EXTERNAL_RESOURCE_USER)
    password = os.getenv(*legion.config.EXTERNAL_RESOURCE_PASSWORD)

    if user and password:
        return requests.auth.HTTPBasicAuth(user, password)

    return None


def copy_file(source_file, target_file):
    """
    Copy file from one location to another

    :param source_file: source file location
    :type source_file: str
    :param target_file: target file location
    :type target_file: str
    :return: None
    """
    shutil.copyfile(source_file, target_file)


def copy_directory_contents(source_directory, target_directory):
    """
    Copy all files from source directory to targer directory

    :param source_directory: source directory
    :type source_directory: str
    :param target_directory: target directory
    :type target_directory: str
    :return: None
    """
    distutils.dir_util.copy_tree(source_directory, target_directory)


def save_file(temp_file, target_file, remove_after_delete=False):
    """
    Upload local file to external resource

    :param temp_file: path to file on local machine
    :type temp_file: str
    :param target_file: path to file on external resource
    :type target_file: str
    :param remove_after_delete: remove local file
    :type remove_after_delete: bool
    :return: str -- path to file on external resource
    """
    try:
        if is_local_resource(target_file):
            if os.path.abspath(temp_file) != os.path.abspath(target_file):
                shutil.copy2(temp_file, target_file)
            result_path = os.path.abspath(target_file)
        else:
            url = normalize_external_resource_path(target_file)

            with open(temp_file, 'rb') as file:
                auth = _get_auth_credentials_for_external_resource()
                response = requests.put(url,
                                        data=file,
                                        auth=auth)
                if response.status_code >= 400:
                    raise Exception('Wrong status code %d returned for url %s' % (response.status_code, url))

            result_path = url

    finally:
        if remove_after_delete:
            os.remove(temp_file)

    return result_path


def download_file(target_file):
    """
    Download file from external resource and return path to file on local machine

    :param target_file: path to file on external resource
    :type target_file: str
    :return: str -- path to file on local machine
    """
    if is_local_resource(target_file):
        return target_file

    # If external resource (only HTTP at this time)
    url = normalize_external_resource_path(target_file)
    name = target_file.split('/')[-1]
    credentials = _get_auth_credentials_for_external_resource()
    response = requests.get(url,
                            stream=True,
                            verify=False,
                            auth=credentials)
    if response.status_code >= 400 or response.status_code < 200:
        raise Exception('Cannot load resource: %s. Returned status code: %d' % (url, response.status_code))

    temp_file = tempfile.mktemp(suffix=name)

    with open(temp_file, 'wb') as file:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                file.write(chunk)

    return os.path.abspath(temp_file)


@contextlib.contextmanager
def extract_archive_item(path, subpath):
    """
    Extract item from archive using context manager to temporary directory

    :param path: path to archive
    :type path: str
    :param subpath: path to file in archive
    :type subpath: str
    :return: str -- path to temporary extracted file
    """
    with TemporaryFolder() as temp_directory:
        try:
            with zipfile.ZipFile(path, 'r') as stream:
                target_path = os.path.join(temp_directory.path, subpath)
                extracted_path = stream.extract(subpath, target_path)

                yield extracted_path
        except zipfile.BadZipFile:
            raise Exception('File {} is not a archive'.format(path))


class ExternalFileReader:
    """
    External file reader for opening files from http://, https:// and local FS
    """

    def __init__(self, path):
        """
        Create external file reader

        :param path: path to file
        :type path: str
        """
        self._path = path
        self._local_path = None
        self._is_external_path = not is_local_resource(self._path)

    @property
    def path(self):
        """
        Get path to file on local machine

        :return:
        """
        return self._local_path

    def _download(self):
        """
        Download external resource

        :return: None
        """
        if not self._local_path:
            if self._is_external_path:
                self._local_path = download_file(self._path)
            else:
                self._local_path = self._path

    def __enter__(self):
        """
        Return self on context enter

        :return: :py:class:`legion.utils.ExternalFileReader`
        """
        self._download()
        return self

    def __exit__(self, exit_type, value, traceback):
        """
        Call remove on context exit

        :param exit_type: -
        :param value: -
        :param traceback: -
        :return: None
        """
        if self._is_external_path:
            remove_directory(self._local_path)


def get_git_revision(file, use_short_hash=True):
    """
    Get current GIT revision of file or directory

    :param file: path to file or directory for check
    :type file: str
    :param use_short_hash: return shorten revision id
    :type use_short_hash: bool
    :return: str or None -- revision id
    """
    try:
        directory = file
        if not os.path.isdir(directory):
            directory = os.path.dirname(file)

        revision = subprocess.check_output(['git', 'rev-parse',
                                            '--short' if use_short_hash else '',
                                            'HEAD'],
                                           cwd=directory)
    except subprocess.CalledProcessError:
        return None

    if isinstance(revision, bytes):
        revision = revision.decode('utf-8')

    return revision.strip()


def send_header_to_stderr(header, value):
    """
    Send header with specific prefix to stderr

    :param header: name of header (without common prefix)
    :type header: str
    :param value: value of header
    :type value: str
    :return: None
    """
    message = '{}{}:{}'.format(legion.containers.headers.STDERR_PREFIX, header, value)
    print(message, file=sys.__stderr__, flush=True)


def model_properties_storage_name(model_id, model_version):
    """
    Construct properties storage name

    :param model_id: model ID
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: str -- name of properties storage
    """
    return 'model-{}-{}'.format(model_id, model_version)


def string_to_bool(value):
    """
    Convert string to bool

    :param value: string or bool
    :type value: str or bool
    :return: bool
    """
    if isinstance(value, bool):
        return value

    return value.lower() in ['true', '1', 't', 'y', 'yes']


def deduce_model_file_name(model_id, model_version):
    """
    Get model file name

    :param model_id: ID of model
    :type model_id: str
    :param model_version: version of model
    :type model_version: str
    :return: str -- auto deduced file name
    """
    date_string = datetime.datetime.now().strftime('%y%m%d%H%M%S')

    valid_user_names = [os.getenv(env) for env in legion.config.MODEL_NAMING_UID_ENV if os.getenv(env)]
    user_id = valid_user_names[0] if valid_user_names else getpass.getuser()

    commit_id = get_git_revision(os.getcwd())
    if not commit_id:
        commit_id = '0000'

    file_name = '%s-%s+%s.%s.%s.model' % (model_id, str(model_version), date_string, user_id, commit_id)

    if string_to_bool(os.getenv(*legion.config.EXTERNAL_RESOURCE_USE_BY_DEFAULT)):
        return '///%s' % file_name

    default_prefix = os.getenv(*legion.config.LOCAL_DEFAULT_RESOURCE_PREFIX)
    if default_prefix:
        return os.path.join(default_prefix, file_name)

    return file_name



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion env names
"""

BUILD_NUMBER = 'BUILD_NUMBER', 0
BUILD_ID = 'BUILD_ID', None
BUILD_URL = 'BUILD_URL', None
BUILD_TAG = 'BUILD_TAG', None

GIT_COMMIT = 'GIT_COMMIT', None
GIT_BRANCH = 'GIT_BRANCH', None

JOB_NAME = 'JOB_NAME', None
NODE_NAME = 'NODE_NAME', None

MODEL_SERVER_URL = 'MODEL_SERVER_URL', 'http://edge'
MODEL_ID = 'MODEL_ID', None
MODEL_FILE = 'MODEL_FILE', None

STATSD_HOST = 'STATSD_HOST', 'graphite'
STATSD_PORT = 'STATSD_PORT', 8125
STATSD_NAMESPACE = 'STATSD_NAMESPACE', 'legion.model'

LEGION_ADDR = 'LEGION_ADDR', '0.0.0.0'
LEGION_PORT = 'LEGION_PORT', 5000

LEGION_API_ADDR = 'LEGION_API_ADDR', '0.0.0.0'
LEGION_API_PORT = 'LEGION_API_PORT', 5001

GRAPHITE_HOST = 'GRAPHITE_HOST', 'graphite'
GRAPHITE_PORT = 'GRAPHITE_PORT', 2003
GRAPHITE_NAMESPACE = 'GRAPHITE_NAMESPACE', 'stats.legion.model'

GRAFANA_URL = 'GRAFANA_URL', 'http://grafana:3000/'
GRAFANA_USER = 'GRAFANA_USER', 'admin'
GRAFANA_PASSWORD = 'GRAFANA_PASSWORD', 'admin'

EDI_URL = 'EDI_URL', 'http://localhost:5001/'
EDI_USER = 'EDI_USER', ''
EDI_PASSWORD = 'EDI_PASSWORD', ''
EDI_TOKEN = 'EDI_TOKEN', ''

LOCAL_DEFAULT_RESOURCE_PREFIX = 'LOCAL_DEFAULT_RESOURCE_PREFIX', ''
EXTERNAL_RESOURCE_USE_BY_DEFAULT = 'EXTERNAL_RESOURCE_USE_BY_DEFAULT', 'true'
EXTERNAL_RESOURCE_PROTOCOL = 'EXTERNAL_RESOURCE_PROTOCOL', 'https'
EXTERNAL_RESOURCE_HOST = 'EXTERNAL_RESOURCE_HOST', 'localhost'
EXTERNAL_RESOURCE_USER = 'EXTERNAL_RESOURCE_USER', None
EXTERNAL_RESOURCE_PASSWORD = 'EXTERNAL_RESOURCE_PASSWORD', None

DOCKER_REGISTRY_USER = 'DOCKER_REGISTRY_USER', None
DOCKER_REGISTRY_PASSWORD = 'DOCKER_REGISTRY_PASSWORD', None

MODEL_NAMING_UID_ENV = 'JUPYTERHUB_USER', 'NB_USER', 'BUILD_ID'
DEBUG = 'DEBUG', 'false'
REGISTER_ON_GRAFANA = 'REGISTER_ON_GRAFANA', 'true'
FLASK_APP_SETTINGS_FILES = 'FLASK_APP_SETTINGS_FILES', None

NAMESPACE = 'NAMESPACE', ''
CLUSTER_CONFIG_PATH = 'CLUSTER_CONFIG_PATH', None
CLUSTER_SECRETS_PATH = 'CLUSTER_SECRETS_PATH', None

MODEL_INSTANCE_SERVICE_ACCOUNT_NAME = 'model'
MODEL_PROPERTIES_CACHE_TTL = 10



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Flask app
"""

import logging
import os
import itertools

import legion.config
import legion.external.grafana
import legion.http
import legion.model
import legion.pymodel
import legion.k8s.properties
import legion.k8s.utils
from flask import Flask, Blueprint, request, jsonify, redirect, render_template
from flask import current_app as app

LOGGER = logging.getLogger(__name__)
blueprint = Blueprint('pyserve', __name__)

SERVE_ROOT = '/'
SERVE_INFO = '/api/model/{model_id}/{model_version}/info'
SERVE_INVOKE = '/api/model/{model_id}/{model_version}/invoke/{endpoint}'
SERVE_INVOKE_DEFAULT = '/api/model/{model_id}/{model_version}/invoke'
SERVE_BATCH = '/api/model/{model_id}/{model_version}/batch/{endpoint}'
SERVE_BATCH_DEFAULT = '/api/model/{model_id}/{model_version}/batch'
SERVE_HEALTH_CHECK = '/healthcheck'

ALL_URLS = SERVE_ROOT, \
           SERVE_INFO, \
           SERVE_INVOKE, SERVE_INVOKE_DEFAULT, \
           SERVE_BATCH, SERVE_BATCH_DEFAULT, \
           SERVE_HEALTH_CHECK


def validate_model_id(model_id, model_version):
    """
    Check that passed model info is equal to current model info

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: None
    """
    if model_id != app.config['model'].model_id:
        raise Exception('Invalid model handler: {}, not {}'.format(app.config['model'].model_id, model_id))
    if model_version != app.config['model'].model_version:
        raise Exception('Invalid model handler: {}, not {}'.format(app.config['model'].model_version, model_version))


@blueprint.route(SERVE_ROOT)
def root():
    """
    Return static file for root query

    :return: :py:class:`Flask.Response` -- response index file
    """
    return redirect('index.html')


@blueprint.route(SERVE_INFO.format(model_id='<model_id>', model_version='<model_version>'))
def model_info(model_id, model_version):
    """
    Get model description

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: :py:class:`Flask.Response` -- model description
    """
    validate_model_id(model_id, model_version)

    model = app.config['model']

    return jsonify(model.description)


@blueprint.route(SERVE_INVOKE_DEFAULT.format(model_id='<model_id>', model_version='<model_version>'),
                 methods=['POST', 'GET'])
@blueprint.route(SERVE_INVOKE.format(model_id='<model_id>', model_version='<model_version>',
                                     endpoint='<endpoint>'), methods=['POST', 'GET'])
def model_invoke(model_id, model_version, endpoint='default'):
    """
    Call model for calculation

    :param model_id: model name
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :param endpoint: target endpoint name
    :type endpoint: str
    :return: :py:class:`Flask.Response` -- result of calculation
    """
    validate_model_id(model_id, model_version)

    input_dict = legion.http.parse_request(request)

    model = app.config['model']
    if endpoint not in model.endpoints:
        raise Exception('Unknown endpoint {!r}'.format(endpoint))

    output = model.endpoints[endpoint].invoke(input_dict)

    return legion.http.prepare_response(output)


@blueprint.route(SERVE_BATCH_DEFAULT.format(model_id='<model_id>', model_version='<model_version>'),
                 methods=['POST'])
@blueprint.route(SERVE_BATCH.format(model_id='<model_id>', model_version='<model_version>', endpoint='<endpoint>'),
                 methods=['POST'])
def model_batch(model_id, model_version, endpoint='default'):
    """
    Call model for calculation in batch mode

    :param model_id: model name
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :param endpoint: target endpoint name
    :type endpoint: str
    :return: :py:class:`Flask.Response` -- result of calculation
    """
    validate_model_id(model_id, model_version)

    input_dicts = legion.http.parse_batch_request(request)

    model = app.config['model']
    if endpoint not in model.endpoints:
        raise Exception('Unknown endpoint {!r}'.format(endpoint))

    responses = [model.endpoints[endpoint].invoke(input_dict) for input_dict in input_dicts]

    return legion.http.prepare_response(responses)


@blueprint.route(SERVE_HEALTH_CHECK)
def healthcheck():
    """
    Check that model is OK

    :return: str -- status string
    """
    return 'OK'


def build_sitemap():
    """
    Build list of valid application URLs

    :return: list[str] -- list of urls
    """
    non_endpoint = [
        url.format(model_id=app.config['model'].model_id,
                   model_version=app.config['model'].model_version)
        for url in ALL_URLS
        if '{endpoint}' not in url
    ]
    with_endpoint = [
        [
            url.format(model_id=app.config['model'].model_id,
                       model_version=app.config['model'].model_version,
                       endpoint=endpoint)
            for url in ALL_URLS
            if '{endpoint}' in url
        ]
        for endpoint in app.config['model'].endpoints.keys()
    ]
    return non_endpoint + list(itertools.chain(*with_endpoint))


def page_not_found_handler(e):
    """
    Exception handler for page not found error

    :param e: NotFound exception
    :type e: :py:class:`werkzeug.exceptions.NotFound`
    :return: tuple[str, int] -- response with error code
    """
    return jsonify(error=True,
                   message=e.description,
                   valid_urls=build_sitemap()), e.code


def init_model(application):
    """
    Load model from app configuration

    :param application: Flask app
    :type application: :py:class:`Flask.app`
    :return: model instance
    """
    if 'MODEL_FILE' not in application.config:
        raise Exception('No model file provided')

    # Select model binary file path
    model_file_path = application.config['MODEL_FILE']
    LOGGER.info('Loading model from {}'.format(model_file_path))

    # Load model container
    model_container = legion.pymodel.Model.load(model_file_path)

    # Load model endpoints
    endpoints = model_container.endpoints  # force endpoints loading
    LOGGER.info('Loaded endpoints: {}'.format(list(endpoints.keys())))

    # Load model properties
    legion.model.set_properties(model_container.properties)

    # Force reload if code run in a cluster and model required any properties
    if legion.k8s.utils.is_code_run_in_cluster() and model_container.required_props:
        legion.model.properties.load()

    return model_container


def create_application():
    """
    Create Flask application and register blueprints

    :return: :py:class:`Flask.app` -- Flask application instance
    """
    static_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'static'))
    application = Flask(__name__, static_url_path='', static_path=static_folder)

    application.register_blueprint(blueprint)

    return application


def register_dashboard(application):
    """
    Register application in Grafana (create dashboard)

    :param application: Flask application instance
    :type application: :py:class:`Flask.app`
    :return: None
    """
    host = os.environ.get(*legion.config.GRAFANA_URL)
    user = os.environ.get(*legion.config.GRAFANA_USER)
    password = os.environ.get(*legion.config.GRAFANA_PASSWORD)

    print('Creating Grafana client for host: %s, user: %s, password: %s' % (host, user, '*' * len(password)))
    client = legion.external.grafana.GrafanaClient(host, user, password)
    client.create_dashboard_for_model(application.config['MODEL_ID'])


def init_application(args=None):
    """
    Initialize configured Flask application instance
    Overall configuration priority: config_default.py, env::FLASK_APP_SETTINGS_FILES file,
    ENV parameters, CLI parameters

    :param args: arguments if provided
    :type args: :py:class:`argparse.Namespace` or None
    :return: :py:class:`Flask.app` -- application instance
    """
    application = create_application()
    legion.http.configure_application(application, args)

    # Put a model object into application configuration
    application.config['model'] = init_model(application)
    application.register_error_handler(404, page_not_found_handler)

    return application


def serve_model(args):
    """
    Serve model

    :param args: arguments
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    logging.info('Legion pyserve initializing')
    application = init_application(args)

    application.run(host=application.config['LEGION_ADDR'],
                    port=application.config['LEGION_PORT'],
                    debug=application.config['DEBUG'],
                    use_reloader=False)

    return application



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Model types for API calls
"""

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

from io import BytesIO
import re

import base64
from PIL import Image as PYTHON_Image
import numpy as np
import pandas as pd

VALID_NATIVE_TYPES = [
    int, float,
    str
]


class BaseType:
    """
    Base model type
    """

    def __init__(self, name=None, native_class=None, description=None, default_numpy_type=np.object):
        """
        Construct base model type

        :param name: name of type
        :type name: str
        :param native_class: type if class builds on native type
        like int, float and etc. (they should be presented in VALID_NATIVE_TYPES)
        :type native_class: int, float and etc or None
        :param description: optional type description
        :type description: str or None
        :param default_numpy_type: default numpy type
        :type :py:class:`numpy.dtype`
        """
        self._name = name
        self._description = description
        self._default_numpy_type = default_numpy_type

        if native_class:
            if native_class not in VALID_NATIVE_TYPES:
                raise Exception('Cannot build type on non-supported native classes')
            self._is_native = True
            self._native_class = native_class
        else:
            self._is_native = False
            self._native_class = native_class

    @property
    def name(self):
        """
        Get name of type

        :return: str
        """
        return self._name if self._name else self.__class__.__name__

    @property
    def description(self):
        """
        Get description of type

        :return: str
        """
        return self._description

    @property
    def default_numpy_type(self):
        """
        Get default numpy type

        :return: :py:class:`numpy.dtype`
        """
        return self._default_numpy_type

    def parse(self, value):
        """
        Parse input value (str, or bytes for files) for base type

        :param value: input value
        :type value: str or bytes
        :return: native representation
        """
        if not self._is_native:
            raise Exception('parse function should be overridden in inherited class')

        return self._native_class(value)

    def export(self, value):
        """
        Export value for base type

        :param value: native representation
        :return: native representation
        """
        if not self._is_native:
            raise Exception('export function should be overridden in inherited class')

        return value

    def __str__(self):
        """
        Get string representation

        :return: str
        """
        return '%s(native_class=%s, description=%s, default_numpy_type=%s)' \
               % (self.name, self._native_class, self._description, self._default_numpy_type)

    def __repr__(self):
        """
        Get string representation

        :return: str
        """
        return '%s(native_class=%r, description=%r, default_numpy_type=%r)' \
               % (self.name, self._native_class, self._description, self._default_numpy_type)


class _Bool(BaseType):
    """
    Type for boolean variables
    """

    TRUE_STRINGS = 'yes', 'true', 't', '1'
    WRONG_STRINGS = 'no', 'false', 'f', '0'

    def __init__(self):
        """
        Construct Bool type
        """
        super(_Bool, self).__init__('Bool', default_numpy_type=np.bool_)

    def parse(self, value):
        """
        Parse boolean strings like 'yes', 'no' and etc. (all listed TRUE_STRINGS in WRONG_STRINGS)

        :param value: input value
        :type value: str or bytes
        :return: bool -- parsed value
        """
        str_value = value.lower()

        if str_value not in self.TRUE_STRINGS + self.WRONG_STRINGS:
            raise ValueError('Invalid value: %s. Valid values is %s and %s' %
                             (value, self.TRUE_STRINGS, self.WRONG_STRINGS))

        return str_value in self.TRUE_STRINGS


class _Image(BaseType):
    """
    Type for images
    """

    BASE64_REGEX = '^data:image/(.+);base64,(.*)$'

    def __init__(self):
        """
        Construct Image type
        """
        super(_Image, self).__init__('Image', default_numpy_type=np.object)

    def parse(self, value):
        """
        Parse image strings or bytes

        :param value: input value
        :type value: str or bytes
        :return: :py:class:`PIL.Image.Image` -- parsed value
        """
        if isinstance(value, str):
            if len(value) > 10 and value[:4] == 'http':
                return self._load_from_network(value)
            elif re.match(self.BASE64_REGEX, value):
                return self._load_from_base64(value)
            else:
                raise Exception('Invalid string: %s' % value)
        elif isinstance(value, bytes):
            file = BytesIO(value)
            return PYTHON_Image.open(file)
        else:
            raise Exception('Invalid data type: %s' % (value.__class__))

    @staticmethod
    def _load_from_network(url):
        """
        Load image from network

        :param url: network image
        :type url: str
        :return: :py:class:`PIL.Image.Image` -- loaded value
        """
        data = urlopen(url).read()
        if isinstance(data, str):
            data = data.encode('ascii')
        file = BytesIO(data)
        img = PYTHON_Image.open(file)
        return img

    def _load_from_base64(self, value):
        """
        Load image from base64

        :param value: base64 string
        :type value: str
        :return: :py:class:`PIL.Image.Image` -- loaded value
        """
        search_results = re.search(self.BASE64_REGEX, value)
        data = search_results.group(2)
        if len(data) % 4:
            data += '=' * (4 - len(data) % 4)
        data = base64.decodebytes(data.encode('ascii'))
        file = BytesIO(data)
        img = PYTHON_Image.open(file)
        return img


Integer = BaseType('Integer', int)
Float = BaseType('Float', float)
String = BaseType('String', str)
Bool = _Bool()
Image = _Image()


# TODO: Rename to LegionType
class ColumnInformation:
    """
    Column information (type, numpy type)
    """

    def __init__(self, representation_type, numpy_type=None):
        """
        Build column information

        :param representation_type: type
        :type representation_type: :py:class:`legion.types.BaseType`
        :param numpy_type: numpy representation type
        :type numpy_type: :py:class:`numpy.dtype` or None
        """
        self._representation_type = representation_type
        self._numpy_type = numpy_type

    @property
    def representation_type(self):
        """
        Get type of column (inherited from BaseType)

        :return: :py:class:`legion.types.BaseType`
        """
        return self._representation_type

    @property
    def numpy_type(self):
        """
        Get numpy type

        :return: :py:class:`numpy.dtype`
        """
        if self._numpy_type:
            return self._numpy_type

        return self.representation_type.default_numpy_type

    @property
    def numpy_type_name(self):
        """
        Get numpy type name

        :return: str -- name of type
        """
        type_instance = self.numpy_type

        if isinstance(type_instance, type):
            return type_instance.__name__

        return type_instance.name

    @property
    def description_for_api(self):
        """
        Get information about columns as a dictionary

        :return: dict[str, any] -- with type and numpy_type columns
        """
        return {
            'type': self.representation_type.name,
            'numpy_type': self.numpy_type_name
        }

    def __repr__(self):
        """
        Get string representation

        :return: str
        """
        return 'ColumnInformation(representation_type=%r, numpy_type=%r)' % (self.representation_type, self.numpy_type)

    def __str__(self):
        """
        Get string representation

        :return: str
        """
        return 'ColumnInformation(representation_type=%s, numpy_type=%s)' % (self.representation_type, self.numpy_type)


# TODO: Add complex
def deduct_types_on_pandas_df(data_frame, extra_columns=None):
    """
    Deduct types on pandas data frame

    :param data_frame: pandas data frame with at least one row
    :type data_frame: :py:class:`pandas.DataFrame`
    :param extra_columns: optional dict
    :type extra_columns: dict[str, :py:class:`legion.types.BaseType`] or None
    :return: dict[str, :py:class:`legion.types.ColumnInformation`]
    """
    types = {}

    for column_name in data_frame.columns:
        column_type = data_frame[column_name].dtype
        sample = data_frame[column_name].iloc[0]

        if isinstance(column_type, np.dtype):
            numpy_name = column_type.name
            if numpy_name[:3] == 'int':
                types[column_name] = ColumnInformation(Integer, column_type)
            elif numpy_name[:4] == 'bool':
                types[column_name] = ColumnInformation(Bool, column_type)
            elif numpy_name[:5] == 'float':
                types[column_name] = ColumnInformation(Float, column_type)
            elif numpy_name[:7] == 'complex':
                types[column_name] = ColumnInformation(Float, column_type)
            elif numpy_name[:6] == 'object':
                if isinstance(sample, str):
                    types[column_name] = ColumnInformation(String, column_type)
                elif isinstance(sample, PYTHON_Image.Image):
                    types[column_name] = ColumnInformation(Image, column_type)
                else:
                    types[column_name] = None
            else:
                types[column_name] = None
        else:
            types[column_name] = None

    if extra_columns:
        for column_name, column_type in extra_columns.items():
            if column_name not in types:
                raise Exception('Undefined column name: %s' % column_name)

            if not isinstance(column_type, BaseType):
                raise Exception('Extra columns dict should contains only instances of classes inherited from BaseType')

            types[column_name] = ColumnInformation(column_type)

    for column_name, column_type in types.items():
        if not column_type:
            raise Exception('Cannot deduct type for column %s' % column_name)

    return types


def build_df(columns_map, input_values, return_dict=False):
    """
    Build pandas.DataFrame (or plain dict) from map of columns and input map of strings or bytes

    :param columns_map: information about columns or None
    :type columns_map: dict[str, :py:class:`legion.types.ColumnInformation`] or None
    :param input_values: input values
    :type input_values: dict[str, union[str, bytes]]
    :param return_dict: return dict instead of pandas DF
    :type return_dict: bool
    :return: :py:class:`pandas.DataFrame` or dict
    """
    values = {}
    types = {}

    if not columns_map:
        return input_values

    for column_name, column_information in columns_map.items():
        if column_name not in input_values:
            raise Exception('Missed value for column %s' % column_name)

        values[column_name] = column_information.representation_type.parse(input_values[column_name])
        types[column_name] = column_information.numpy_type

    if return_dict:
        return values

    data_frame = pd.DataFrame([values])
    data_frame = data_frame.astype(types)
    return data_frame


def get_column_types(param_types):
    """
    Build dict with ColumnInformation from param_types argument for export function

    :param param_types: pandas DF with custom dict or pandas DF.
    Custom dict contains of column_name => legion.BaseType
    :type param_types tuple(:py:class:`pandas.DataFrame`, dict) or :py:class:`pandas.DataFrame`
    :return: dict[str, :py:class:`legion.types.ColumnInformation`] -- column name => column information
    """
    custom_props = None

    if isinstance(param_types, tuple) and len(param_types) == 2 \
            and isinstance(param_types[0], pd.DataFrame) \
            and isinstance(param_types[1], dict):

        pandas_df_sample = param_types[0]
        custom_props = param_types[1]
    elif isinstance(param_types, pd.DataFrame):
        pandas_df_sample = param_types
    else:
        raise Exception('Provided invalid param types: not tuple[DataFrame, dict] or DataFrame')

    return deduct_types_on_pandas_df(data_frame=pandas_df_sample, extra_columns=custom_props)


def deduce_param_types(data_frame, optional_dictionary=None):
    """
    Deduce param types of pandas DF. Optionally overwrite to custom legion.BaseType

    :param data_frame: pandas DF
    :type data_frame: :py:class:`pandas.DataFrame`
    :param optional_dictionary: custom dict contains of column_name => legion.types.BaseType
    :type optional_dictionary: dict[str, :py:class:`legion.types.BaseType`]
    :return: dict[str, :py:class:`legion.types.ColumnInformation`]
    """
    if optional_dictionary:
        return _get_column_types((data_frame, optional_dictionary))

    return _get_column_types(data_frame)


int8 = ColumnInformation(Integer, np.int8)
uint8 = ColumnInformation(Integer, np.uint8)
int16 = ColumnInformation(Integer, np.int16)
uint16 = ColumnInformation(Integer, np.uint16)
int32 = ColumnInformation(Integer, np.int32)
uint32 = ColumnInformation(Integer, np.uint32)
int64 = ColumnInformation(Integer, np.int64)
uint64 = ColumnInformation(Integer, np.uint64)

float16 = ColumnInformation(Float, np.float16)
float32 = ColumnInformation(Float, np.float32)
float64 = ColumnInformation(Float, np.float64)

string = ColumnInformation(String)
boolean = ColumnInformation(Bool)
image = ColumnInformation(Image)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Python model
"""
import json
import sys
import os
import zipfile

import logging
import legion.config
import legion.containers.headers
import legion.model
import legion.k8s.properties
import legion.model.types
import legion.metrics
from legion.utils import model_properties_storage_name, send_header_to_stderr, \
    extract_archive_item, TemporaryFolder, deduce_model_file_name, save_file


import dill


LOGGER = logging.getLogger(__name__)


class ModelEndpoint:
    """
    Object describes named model invocation endpoint
    """

    def __init__(self, *, name, apply, column_types, prepare, use_df):
        """
        Build model endpoint

        :param name: endpoint name
        :type name: str
        :param apply: apply function (callable object)
        :type apply: callable
        :param column_types: (Optional) column types dictionary
        :type column_types: dict[str, ColumnInformation]
        :param prepare: (Optional) prepare function (callable object)
        :type prepare: callable
        :param use_df: use db for prepare function
        :type use_df: bool
        """
        self._name = name
        self._apply = apply
        self._column_types = column_types
        self._prepare = prepare
        self._use_df = use_df

    @property
    def description(self):
        """
        Get model description

        :return: dict[str, any] -- dict with model description
        """
        data = {
            'name': self.name,
            'use_df': self.use_df
        }

        if self.column_types:
            data['input_params'] = {k: v.description_for_api for (k, v) in self.column_types.items()}
        else:
            data['input_params'] = False

        return data

    def invoke(self, input_vector):
        """
        Calculate result of model execution

        :param input_vector: input data
        :type input_vector: dict[str, union[str, Image]]
        :return: dict -- output data
        """
        LOGGER.info('Input vector: %r' % input_vector)
        data_frame = legion.model.types.build_df(self.column_types, input_vector, not self.use_df)

        LOGGER.info('Running prepare with DataFrame: %r' % data_frame)
        data_frame = self.prepare(data_frame)

        LOGGER.info('Applying function with DataFrame: %s' % str(data_frame))
        response = self.apply(data_frame)
        LOGGER.info('Returning response: %s' % str(response))

        return response

    @property
    def name(self):
        """
        Get endpoint name

        :return: str -- name of endpoint
        """
        return self._name

    @property
    def apply(self):
        """
        Get apply function

        :return: function
        """
        return self._apply

    @property
    def column_types(self):
        """
        Get column types

        :return:
        """
        return self._column_types

    @property
    def prepare(self):
        """
        Get prepare function

        :return:
        """
        return self._prepare

    @property
    def use_df(self):
        """
        Use pandas DataFrame for input parameters

        :return: bool -- use DataFrame
        """
        return self._use_df

    def __str__(self):
        """
        Get string representation

        :return: str -- string representation
        """
        return 'ModelEndpoint for endpoint {}'.format(self.name)

    __repr__ = __str__


class Model:
    NAME = 'pymodel'

    ZIP_COMPRESSION = zipfile.ZIP_STORED
    ZIP_FILE_MODEL = 'model'
    ZIP_FILE_INFO = 'manifest.json'
    ZIP_FILE_PROPERTIES = 'properties'

    PROPERTY_MODEL_ID = 'model.id'
    PROPERTY_MODEL_VERSION = 'model.version'
    PROPERTY_ENDPOINT_NAMES = 'model.endpoints'
    PROPERTY_REQUIRED_PROPERTIES = 'model.required_properties'

    def __init__(self, model_id, model_version):
        """
        Build empty model container

        :param model_id: model ID
        :type model_id: str
        :param model_version: model version
        :type model_version: str
        """
        self._meta_information = {}  # type: dict

        self._model_id = model_id
        self._model_version = model_version

        self._required_properties = []

        self._endpoints = {}  # type: dict or None
        self._path = None  # type: str or None

        storage_name = model_properties_storage_name(self.model_id, self.model_version)
        self._properties = legion.k8s.K8SConfigMapStorage(storage_name,
                                                          cache_ttl=legion.config.MODEL_PROPERTIES_CACHE_TTL)

        send_header_to_stderr(legion.containers.headers.MODEL_ID, self.model_id)
        send_header_to_stderr(legion.containers.headers.MODEL_VERSION, self.model_version)

    @property
    def properties(self):
        """
        Get model properties

        :return:
        """
        return self._properties

    def _load_from_archive(self, path):
        """
        Populate model container with data from file after initialization

        :param path: path to model binary
        :type path: str
        :return: None
        """
        self._path = path
        self._endpoints = None

        with extract_archive_item(path, Model.ZIP_FILE_INFO) as manifest_path:
            with open(manifest_path, 'r') as manifest_file:
                self._meta_information = json.load(manifest_file)

        self._required_properties = self._meta_information[self.PROPERTY_REQUIRED_PROPERTIES]

        with extract_archive_item(path, Model.ZIP_FILE_PROPERTIES) as properties_path:
            with open(properties_path, 'r') as properties_file:
                self.properties.data = json.load(properties_file)

    @staticmethod
    def load(path):
        """
        Populate model container with data from file and initialize

        :param path: path to model binary
        :type path: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        if not os.path.exists(path):
            raise Exception('File not existed: {}'.format(path))

        with extract_archive_item(path, Model.ZIP_FILE_INFO) as manifest_path:
            with open(manifest_path, 'r') as manifest_file:
                manifest_data = json.load(manifest_file)

                model_id = manifest_data[Model.PROPERTY_MODEL_ID]
                model_version = manifest_data[Model.PROPERTY_MODEL_VERSION]

        instance = Model(model_id, model_version)
        instance._load_from_archive(path)

        return instance

    @staticmethod
    def _build_endpoint_file_name(endpoint_name):
        """
        Build name of endpoint file in model binary

        :param endpoint_name: endpoint name
        :type endpoint_name: str
        :return: inner path to file
        :rtype: str
        """
        return endpoint_name

    def load_endpoint(self, endpoint_name):
        """
        Load endpoint from model binary

        :param endpoint_name: endpoint name
        :type endpoint_name: str
        :return: deserialized model endpoint
        :rtype: :py:class:`legion.pymodel.model.ModelEndpoint`
        """
        with extract_archive_item(self._path, self._build_endpoint_file_name(endpoint_name)) as endpoint_path:
            with open(endpoint_path, 'rb') as endpoint_file:
                return dill.load(endpoint_file)

    @property
    def description(self):
        """
        Get model container description

        :return: dict[str, any] -- model description
        """
        return {
            'model_id': self.model_id,
            'model_version': self.model_version,
            'endpoints': {ep.name: ep.description for ep in self.endpoints.values()}
        }

    @property
    def endpoints(self):
        """
        Get or lazy load endpoints

        :return: dict -- current endpoints
        """
        # If endpoint is None it means that model in partial loaded state
        # (properties has been loaded, endpoints - not)
        if self._endpoints is None:
            self._endpoints = {}

            endpoint_names = self._meta_information.get(self.PROPERTY_ENDPOINT_NAMES)
            if not endpoint_names:
                raise Exception('PyModel does not contain {} field or field is empty'
                                .format(self.PROPERTY_ENDPOINT_NAMES))

            for endpoint_name in endpoint_names:
                self._endpoints[endpoint_name] = self.load_endpoint(endpoint_name)

        return self._endpoints

    def save(self, path=None):
        """
        Save model to path (or deduce path)

        :param path: (Optional) target save name
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        if not self.endpoints:
            raise ValueError('Cannot save empty model container (no one export function has been called)')

        meta_information_to_save = self._meta_information.copy()
        meta_information_to_save.update(self._collect_build_info())

        # Add model id, model version, model endpoint
        meta_information_to_save[self.PROPERTY_MODEL_ID] = self.model_id
        meta_information_to_save[self.PROPERTY_MODEL_VERSION] = self.model_version
        meta_information_to_save[self.PROPERTY_ENDPOINT_NAMES] = list(self._endpoints.keys())
        meta_information_to_save[self.PROPERTY_REQUIRED_PROPERTIES] = list(self._required_properties)

        self._path = path

        file_name_has_been_deduced = False
        if not self._path:
            self._path = deduce_model_file_name(self.model_id, self.model_version)

        # Save
        with TemporaryFolder('legion-model-save') as temp_directory:
            temp_file = os.path.join(temp_directory.path, 'result.zip')
            with zipfile.ZipFile(temp_file, 'w', self.ZIP_COMPRESSION) as stream:
                # Add manifest file
                with open(os.path.join(temp_directory.path, self.ZIP_FILE_INFO), 'w') as info_file:
                    json.dump(meta_information_to_save, info_file)
                stream.write(os.path.join(temp_directory.path, self.ZIP_FILE_INFO), self.ZIP_FILE_INFO)

                # Add current properties state
                with open(os.path.join(temp_directory.path, self.ZIP_FILE_PROPERTIES), 'w') as props_file:
                    json.dump(legion.model.properties.data, props_file)
                stream.write(os.path.join(temp_directory.path, self.ZIP_FILE_PROPERTIES), self.ZIP_FILE_PROPERTIES)

                # Add endpoints
                for endpoint in self.endpoints.values():
                    path = self._build_endpoint_file_name(endpoint.name)
                    with open(os.path.join(temp_directory.path, path), 'wb') as endpoint_file:
                        dill.dump(endpoint, endpoint_file, recurse=True)
                    stream.write(os.path.join(temp_directory.path, path), path)

            result_path = save_file(temp_file, self._path)

        send_header_to_stderr(legion.containers.headers.MODEL_PATH, result_path)
        send_header_to_stderr(legion.containers.headers.SAVE_STATUS, 'OK')

        if file_name_has_been_deduced:
            print('Model has been saved to {}'.format(result_path), file=sys.stderr)

        return self

    def _export(self, apply_func, prepare_func, column_types, use_df, endpoint_name):
        """
        Export simple Pandas based model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param column_types: result of deduce_param_types or prepared column information or None
        :type column_types: dict[str, :py:class:`legion.model.types.ColumnInformation`] or None
        :param use_df: use pandas DF for prepare and apply function
        :type use_df: bool
        :param endpoint_name: name of endpoint
        :type endpoint_name: str
        :return: None
        """
        if not hasattr(apply_func, '__call__'):
            raise Exception('Provided non-callable object as apply_function')

        if column_types:
            if not isinstance(column_types, dict) \
                    or not column_types.keys() \
                    or not isinstance(list(column_types.values())[0], legion.model.types.ColumnInformation):
                raise Exception('Bad param_types / input_data_frame provided')

        if prepare_func is None:
            def prepare_func(input_dict):
                """
                Return input value (default prepare function)
                :param x: dict of values
                :return: dict of values
                """
                return input_dict

        if endpoint_name in self._endpoints:
            raise Exception('Endpoint {} already existed in endpoints'.format(endpoint_name))

        new_endpoint = ModelEndpoint(name=endpoint_name,
                                     apply=apply_func,
                                     column_types=column_types,
                                     prepare=prepare_func,
                                     use_df=use_df)

        self.endpoints[endpoint_name] = new_endpoint

    def export_df(self, apply_func, input_data_frame, *, prepare_func=None, endpoint='default'):
        """
        Export simple Pandas DF based model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param input_data_frame: pandas DF
        :type input_data_frame: :py:class:`pandas.DataFrame`
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        column_types = legion.model.types.get_column_types(input_data_frame)
        self._export(apply_func, prepare_func, column_types, True, endpoint)
        return self

    def export(self, apply_func, column_types, *, prepare_func=None, endpoint='default'):
        """
        Export simple parameters defined model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param column_types: result of deduce_param_types or prepared column information
        :type column_types: dict[str, :py:class:`legion.model.types.ColumnInformation`]
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._export(apply_func, prepare_func, column_types, False, endpoint)
        return self

    def export_untyped(self, apply_func, *, prepare_func=None, endpoint='default'):
        """
        Export simple untyped model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._export(apply_func, prepare_func, None, False, endpoint)
        return self

    @property
    def model_id(self):
        """
        Get model id

        :return: str or None -- model id
        """
        return self._model_id

    @property
    def model_version(self):
        """
        Get model version

        :return: str or None -- model version
        """
        return self._model_version

    @property
    def required_props(self):
        """
        Get model required props

        :return: list[str] -- list of required property names
        """
        return self._required_properties

    @property
    def meta_information(self):
        """
        Get copy of meta information

        :return: dict -- copy of meta information
        """
        return self._meta_information.copy()

    def _collect_build_info(self):
        """
        Get additional container properties for container

        :return: dict[str, any] -- additional container properties
        """
        return {
            'legion.version': legion.__version__,

            'jenkins.build_number': os.environ.get(*legion.config.BUILD_NUMBER),
            'jenkins.build_id': os.environ.get(*legion.config.BUILD_ID),
            'jenkins.build_tag': os.environ.get(*legion.config.BUILD_TAG),
            'jenkins.build_url': os.environ.get(*legion.config.BUILD_URL),

            'jenkins.git_commit': os.environ.get(*legion.config.GIT_COMMIT),
            'jenkins.git_branch': os.environ.get(*legion.config.GIT_BRANCH),

            'jenkins.node_name': os.environ.get(*legion.config.NODE_NAME),
            'jenkins.job_name': os.environ.get(*legion.config.JOB_NAME)
        }

    def send_metric(self, metric, value):
        """
        Send build metric value

        :param metric: metric type or metric name
        :type metric: :py:class:`legion.metrics.Metric` or str
        :param value: metric value
        :type value: float or int
        :return: None
        """
        return legion.metrics.send_metric(self.model_id, metric, value)

    def define_property(self, name, initial_value):
        """
        Define model property and set initial value

        :param name: property name
        :type name: str
        :param initial_value: initial property value
        :type initial_value: any
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._required_properties.append(name)

        if not legion.model.properties:
            raise Exception('Model properties has not been initialized')

        legion.model.properties[name] = initial_value

        return self



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
IO Notifier module for Legion Template
"""
import logging
import os

import aionotify
import yaml

LOGGER = logging.getLogger(__name__)


async def file_change_monitor(template_system, filepath, is_yaml=False, var_name='file'):
    """
     On file change updates template context with a file content
     and renders it.

     Example #1 (Yaml file) :
        {{ load_module('legion.ltemplate.render', filepath='config.yml', is_yaml=True, var_name='conf') }}
        <b>Hosts values:</b>
        <ul>
        {% for item in conf:  %}
        <li>{{ item.hosts }}</li>
        {% endfor %}
        </ul>

    Example #2 (Txt file) :
        {{ load_module('legion.io.render', filepath='config.txt') }}
        <h1>Config file Context:</h1>
        <pre>{{ file }}</pre>

    :param template_system: Object, that contains 'render' callback function
    :param filepath: Path to a File, relative to template file or absolute
    :type filepath: str
    :param is_yaml_file: Indicates if file is in yaml format and it should be parsed
    :type is_yaml_file: bool
    :param var_name: variable name, used to store file content in context
    :type var_name: str
    :return: None
    """
    if not os.path.isabs(filepath):
        filepath = os.path.join(os.path.dirname(template_system.template_file_path), filepath)

    LOGGER.info('Starting file change monitor for file {}, is_yaml={} as variable={}'
                .format(filepath, is_yaml, var_name))

    with open(filepath) as f:
        if is_yaml:
            template_system.render(**{var_name: yaml.load(f)})
        else:
            template_system.render(**{var_name: f.read()})

    LOGGER.debug('Initializing watcher for file {}'.format(filepath))
    watcher = aionotify.Watcher()
    watcher.watch(path=filepath, flags=aionotify.Flags.MODIFY)

    LOGGER.debug('Awaiting setup of watcher for file {}'.format(filepath))
    await watcher.setup(template_system._loop)

    LOGGER.debug('Starting aionotify loop')
    while True:
        await watcher.get_event()

        LOGGER.info('Got update event for file {}'.format(filepath))
        with open(filepath) as f:
            if is_yaml:
                template_system.render(**{var_name: yaml.load(f)})
            else:
                template_system.render(**{var_name: f.read()})



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s functions
"""
import logging
import os
import json

import docker
import docker.errors

import legion
import legion.model
import legion.pymodel
import legion.config
import legion.containers.headers
import legion.utils

LOGGER = logging.getLogger(__name__)


def build_docker_client(args=None):
    """
    Create docker client

    :param args: command arguments
    :type args: :py:class:`argparse.Namespace` or None
    :return: :py:class:`docker.Client`
    """
    client = docker.from_env()
    return client


def get_docker_container_id_from_cgroup_line(line):
    """
    Get docker container id from proc cgroups line

    :argument line: line from /proc/<pid>/cgroup
    :type line: str
    :return: str -- container ID
    """
    parts = line.split('/')

    try:
        if 'docker' in parts:
            docker_pos = parts.index('docker')
            return parts[docker_pos + 1]
        elif 'kubepods' in parts:
            kubepods_pos = parts.index('kubepods')
            return parts[kubepods_pos + 3]
        else:
            raise Exception('Cannot find docker or kubepods tag in cgroups')
    except Exception as container_detection_error:
        raise Exception('Cannot find container ID in line {}: {}'.format(line, container_detection_error))


def get_current_docker_container_id():
    """
    Get ID of current docker container using proc cgroups

    :return: str -- current container id
    """
    with open('/proc/self/cgroup') as f:
        lines = [line.strip('\n') for line in f]
        longest_line = max(lines, key=len)
        return get_docker_container_id_from_cgroup_line(longest_line)


def commit_image(client, container_id=None):
    """
    Commit container and return image sha commit id

    :param client: Docker client
    :type client: :py:class:`docker.client.DockerClient`
    :param container_id: (Optional) id of target container. Current if None
    :type container_id: str
    :return: str -- docker image id
    """
    if not container_id:
        container_id = get_current_docker_container_id()

    container = client.containers.get(container_id)
    image = container.commit()

    return image.id


def build_docker_image(client, model_id, model_file, labels,
                       docker_image_tag):
    """
    Build docker image from current image with addition files

    :param client: Docker client
    :type client: :py:class:`docker.client.DockerClient`
    :param model_id: model id
    :type model_id: str
    :param model_file: path to model file (in the temporary directory)
    :type model_file: str
    :param labels: image labels
    :type labels: dict[str, str]
    :param docker_image_tag: str docker image tag
    :type docker_image_tag: str ot None
    :return: docker.models.Image
    """
    with legion.utils.TemporaryFolder('legion-docker-build') as temp_directory:
        # Copy Jenkins workspace from mounted volume to persistent container directory
        target_workspace = '/app'
        target_model_file = os.path.join(target_workspace, model_id)

        try:
            LOGGER.info('Copying model binary from {!r} to {!r}'
                        .format(model_file, target_model_file))
            legion.utils.copy_file(model_file, target_model_file)
        except Exception as model_binary_copy_exception:
            LOGGER.exception('Unable to move model binary to persistent location',
                             exc_info=model_binary_copy_exception)
            raise

        try:
            workspace_path = os.getcwd()
            LOGGER.info('Copying model workspace from {!r} to {!r}'.format(workspace_path, target_workspace))
            legion.utils.copy_directory_contents(workspace_path, target_workspace)
        except Exception as model_workspace_copy_exception:
            LOGGER.exception('Unable to move model workspace to persistent location',
                             exc_info=model_workspace_copy_exception)
            raise

        # Copy additional files for docker build
        additional_directory = os.path.abspath(os.path.join(
            os.path.dirname(__file__),
            '..', 'templates', 'docker_files'
        ))
        legion.utils.copy_directory_contents(additional_directory, temp_directory.path)

        # ALL Filesystem modification below next line would be ignored
        captured_image_id = commit_image(client)
        LOGGER.info('Image {} has been captured'.format(captured_image_id))

        docker_file_content = legion.utils.render_template('Dockerfile.tmpl', {
            'DOCKER_BASE_IMAGE_ID': captured_image_id,
            'MODEL_ID': model_id,
            'MODEL_FILE': target_model_file
        })

        labels = {k: str(v) if v else None
                  for (k, v) in labels.items()}

        with open(os.path.join(temp_directory.path, 'Dockerfile'), 'w') as file:
            file.write(docker_file_content)

        LOGGER.info('Building docker image in folder {}'.format(temp_directory.path))
        logs = None
        try:
            image, logs = client.images.build(
                tag=docker_image_tag,
                nocache=True,
                path=temp_directory.path,
                rm=True,
                labels=labels
            )
        except Exception as build_error:
            LOGGER.error('Cannot build image: {}. Build logs: {}'.format(build_error, logs))
            raise

        return image


def generate_docker_labels_for_image(model_file, model_id, args):
    """
    Generate docker image labels from model file

    :param model_file: path to model file
    :type model_file: str
    :param model_id: model id
    :type model_id: str
    :param args: command arguments
    :type args: :py:class:`argparse.Namespace`
    :return: dict[str, str] of labels
    """
    container = legion.pymodel.Model.load(model_file)

    base = {
        legion.containers.headers.DOMAIN_MODEL_ID: model_id,
        legion.containers.headers.DOMAIN_MODEL_VERSION: container.model_version,
        legion.containers.headers.DOMAIN_CLASS: 'pyserve',
        legion.containers.headers.DOMAIN_CONTAINER_TYPE: 'model',
        legion.containers.headers.DOMAIN_MODEL_PROPERTIES: ','.join(container.required_props),
        legion.containers.headers.DOMAIN_MODEL_PROPERTY_VALUES: json.dumps(container.properties.data)
    }

    for key, value in container.meta_information.items():
        if hasattr(value, '__iter__') and not isinstance(value, str):
            formatted_value = ', '.join(item for item in value)
        else:
            formatted_value = str(value)

        base[legion.containers.headers.DOMAIN_PREFIX + key] = formatted_value

    return base


def generate_docker_labels_for_container(image):
    """
    Build container labels from image labels (copy)

    :param image: source Docker image
    :type image: :py:class:`docker.models.image.Image`
    :return: dict[str, str] of labels
    """
    return image.labels


def push_image_to_registry(client, image, external_image_name):
    """
    Push docker image to registry

    :param client: Docker client
    :type client: :py:class:`docker.client.DockerClient`
    :param image: Docker image
    :type image: :py:class:`docker.models.images.Image`
    :param external_image_name: target Docker image name (with repository)
    :type external_image_name: str
    :return: None
    """
    docker_registry = external_image_name
    version = None

    registry_delimiter = docker_registry.find('/')
    if registry_delimiter < 0:
        raise Exception('Invalid registry format')

    registry = docker_registry[:registry_delimiter]
    image_name = docker_registry[registry_delimiter + 1:]

    version_delimiter = image_name.find(':')
    if version_delimiter > 0:
        version = image_name[version_delimiter + 1:]
        image_name = image_name[:version_delimiter]

    docker_registry_user = os.getenv(*legion.config.DOCKER_REGISTRY_USER)
    docker_registry_password = os.getenv(*legion.config.DOCKER_REGISTRY_PASSWORD)
    auth_config = None

    if docker_registry_user and docker_registry_password:
        auth_config = {
            'username': docker_registry_user,
            'password': docker_registry_password
        }

    image_and_registry = '{}/{}'.format(registry, image_name)

    image.tag(image_and_registry, version)
    LOGGER.info('Pushing {}:{} to {}'.format(image_and_registry, version, registry))

    client.images.push(image_and_registry, tag=version, auth_config=auth_config)
    LOGGER.info('Successfully pushed image {}:{}'.format(image_and_registry, version))

    image_with_version = '{}/{}:{}'.format(registry, image_name, version)
    legion.utils.send_header_to_stderr(legion.containers.headers.IMAGE_TAG_EXTERNAL, image_with_version)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Graphana API functional for working with models
"""

import json
import os

import legion.config
from legion.utils import render_template

import requests


class GrafanaClient:
    """
    Base Grafana HTTP API client
    """

    def __init__(self, base, user=None, password=None):
        """
        Build client

        :param base: base url, for example: http://parallels/grafana/
        :type base: str
        :param user: user name
        :type user: str or None
        :param password: user password
        :type password: str or None
        """
        self._base = base.strip('/')
        self._user = user
        self._password = password

    def _query(self, url, payload=None, action='GET'):
        """
        Perform query to Grafana server

        :param url: query suburl, for example: /api/search/
        :type url: str
        :param payload: payload (will be converted to JSON) or None
        :type payload: dict[str, any]
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :type action: str
        :return: dict[str, any] -- response content
        """
        full_url = self._base + url

        headers = {
            'Content-Type': 'application/json'
        }

        auth = None
        if self._user and self._password:
            auth = (self._user, self._password)

        response = requests.request(action.lower(), full_url, json=payload, headers=headers, auth=auth)

        if response.status_code in (401, 403):
            raise Exception('Auth failed')

        if response.status_code != 200:
            raise Exception('Wrong answer for url = %s: %s' % (full_url, repr(response)))

        answer = json.loads(response.text)

        return answer

    def delete_dashboard(self, dashboard_uri):
        """
        Delete dashboard by url

        :param dashboard_uri: dashboard uri
        :type dashboard_uri: str
        :return: None
        """
        self._query('/api/dashboards/%s' % dashboard_uri, action='DELETE')

    def remove_dashboard_for_model(self, model_id, model_version):
        """
        Remove model's dashboard

        :param model_id: model id
        :type model_id: str
        :param model_version: model version
        :type model_id: str or None
        :return: None
        """
        if self.is_dashboard_exists(model_id, model_version):
            dashboard = self.get_model_dashboard(model_id, model_version)
            self.delete_dashboard(dashboard['uri'])

    def is_dashboard_exists(self, model_id, model_version):
        """
        Check if model's dashboard exists

        :param model_id: model id
        :type model_id: str
        :param model_version: model version
        :type model_id: str or None
        :return: bool -- is dashboard exists
        """
        return self.get_model_dashboard(model_id, model_version) is not None

    def get_model_dashboard(self, model_id, model_version):
        """
        Search for model's dashboard

        :param model_id: model id
        :type model_id: str
        :param model_version: model version
        :type model_id: str or None
        :return: dict with dashboard information or None
        """
        data = self._query('/api/search/?tag=model_{}'.format(model_id))
        if not data:
            return None

        return data[0]

    def create_dashboard_for_model(self, model_id, model_version=None):
        """
        Create model's dashboard from template

        :param model_id: model id
        :type model_id: str
        :param model_version: model version
        :type model_id: str or None
        :return: None
        """
        self.remove_dashboard_for_model(model_id, model_version)

        json_string = render_template('grafana-dashboard.json.tmpl', {
            'MODEL_ID': model_id,
            'MODEL_VERSION': model_version
        })

        dashboard = json.loads(json_string)

        payload = {
            'overwrite': False,
            'dashboard': dashboard
        }
        self._query('/api/dashboards/db', payload, 'POST')


def build_client(args):
    """
    Build Grafana client from ENV and from command line arguments

    :param args: command arguments
    :type args: :py:class:`argparse.Namespace`
    :return: :py:class:`legion.grafana.GrafanaClient`
    """
    host = os.environ.get(*legion.config.GRAFANA_URL)
    user = os.environ.get(*legion.config.GRAFANA_USER)
    password = os.environ.get(*legion.config.GRAFANA_PASSWORD)

    if args.grafana_server:
        host = args.grafana_server

    if args.grafana_user:
        user = args.grafana_user

    if args.grafana_password:
        password = args.grafana_password

    client = GrafanaClient(host, user, password)

    return client
